"""Main window for the Lutris interface."""

# pylint: disable=too-many-lines
# pylint: disable=no-member
import os
from collections import namedtuple
from datetime import datetime
from gettext import gettext as _
from gettext import ngettext
from typing import Dict, Iterable, List, Set, cast
from urllib.parse import unquote, urlparse

from gi.repository import Gdk, Gio, GLib, Gtk

from lutris import services, settings
from lutris.api import (
    LUTRIS_ACCOUNT_CONNECTED,
    LUTRIS_ACCOUNT_DISCONNECTED,
    get_runtime_versions,
)
from lutris.database import categories
from lutris.database import categories as categories_db
from lutris.database import games as games_db
from lutris.database import saved_searches as saved_searches_db
from lutris.database.categories import CATEGORIES_UPDATED
from lutris.database.saved_searches import SAVED_SEARCHES_UPDATED
from lutris.database.services import ServiceGameCollection
from lutris.exceptions import EsyncLimitError, InvalidSearchTermError
from lutris.game import GAME_INSTALLED, GAME_STOPPED, GAME_UNHANDLED_ERROR, GAME_UPDATED, Game
from lutris.gui import dialogs
from lutris.gui.addgameswindow import AddGamesWindow
from lutris.gui.config.edit_saved_search import SearchFiltersBox
from lutris.gui.config.preferences_dialog import PreferencesDialog
from lutris.gui.dialogs import ClientLoginDialog, ErrorDialog, QuestionDialog, get_error_handler, register_error_handler
from lutris.gui.dialogs.delegates import DialogInstallUIDelegate, DialogLaunchUIDelegate
from lutris.gui.dialogs.game_import import ImportGameDialog
from lutris.gui.download_queue import DownloadQueue
from lutris.gui.views import (
    COL_INSTALLED_AT,
    COL_INSTALLED_AT_TEXT,
    COL_LASTPLAYED,
    COL_LASTPLAYED_TEXT,
    COL_NAME,
    COL_PLAYTIME,
    COL_PLAYTIME_TEXT,
    COL_SORTNAME,
    COL_YEAR,
)
from lutris.gui.views.grid import GameGridView
from lutris.gui.views.list import GameListView
from lutris.gui.views.store import GameStore
from lutris.gui.widgets.game_bar import GameBar
from lutris.gui.widgets.gi_composites import GtkTemplate
from lutris.gui.widgets.sidebar import LutrisSidebar, SidebarRow
from lutris.gui.widgets.utils import load_icon_theme, open_uri, pick_stock_icon
from lutris.runtime import ComponentUpdater, RuntimeUpdater
from lutris.search import GameSearch
from lutris.search_predicate import NotPredicate
from lutris.services.base import SERVICE_GAMES_LOADED, SERVICE_LOGIN, SERVICE_LOGOUT
from lutris.services.lutris import LutrisService, sync_media
from lutris.style_manager import THEME_CHANGED
from lutris.util import datapath
from lutris.util.busy import BUSY_STARTED, BUSY_STOPPED
from lutris.util.jobs import COMPLETED_IDLE_TASK, AsyncCall, schedule_at_idle
from lutris.util.library_sync import LOCAL_LIBRARY_UPDATED, LibrarySyncer
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.path_cache import MISSING_GAMES, add_to_path_cache
from lutris.util.strings import get_natural_sort_key
from lutris.util.system import update_desktop_icons
from lutris.util.wine.wine import clear_wine_version_cache


@GtkTemplate(ui=os.path.join(datapath.get(), "ui", "lutris-window.ui"))
class LutrisWindow(Gtk.ApplicationWindow, DialogLaunchUIDelegate, DialogInstallUIDelegate):  # type:ignore[misc]
    """Handler class for main window signals."""

    default_view_type = "grid"
    default_width = 800
    default_height = 600

    __gtype_name__ = "LutrisWindow"
    games_stack: Gtk.Stack = GtkTemplate.Child()
    sidebar_revealer: Gtk.Revealer = GtkTemplate.Child()
    sidebar_scrolled: Gtk.ScrolledWindow = GtkTemplate.Child()
    game_revealer: Gtk.Revealer = GtkTemplate.Child()
    search_entry: Gtk.SearchEntry = GtkTemplate.Child()
    search_filters_button: Gtk.MenuButton = GtkTemplate.Child()
    search_box: Gtk.Box = GtkTemplate.Child()
    zoom_adjustment: Gtk.Adjustment = GtkTemplate.Child()
    blank_overlay: Gtk.Alignment = GtkTemplate.Child()
    viewtype_icon: Gtk.Image = GtkTemplate.Child()
    download_revealer: Gtk.Revealer = GtkTemplate.Child()
    game_view_spinner: Gtk.Spinner = GtkTemplate.Child()
    login_notification_revealer: Gtk.Revealer = GtkTemplate.Child()
    lutris_log_in_label: Gtk.Label = GtkTemplate.Child()
    version_notification_revealer: Gtk.Revealer = GtkTemplate.Child()
    version_notification_label: Gtk.Revealer = GtkTemplate.Child()
    show_hidden_games_button: Gtk.ModelButton = GtkTemplate.Child()

    def __init__(self, application=None, **kwargs) -> None:
        width = int(settings.read_setting("width") or self.default_width)
        height = int(settings.read_setting("height") or self.default_height)
        super().__init__(
            default_width=width,
            default_height=height,
            window_position=Gtk.WindowPosition.NONE,
            name="lutris",
            icon_name="net.lutris.Lutris",
            application=application,
            **kwargs,
        )
        update_desktop_icons()
        load_icon_theme()
        self.set_wmclass("net.lutris.Lutris", "net.lutris.Lutris")
        self.application = application
        self.window_x, self.window_y = self.get_position()
        self.restore_window_position()
        self.threads_stoppers = []
        self.window_size = (width, height)
        self.maximized = settings.read_setting("maximized") == "True"
        self.service = None
        self.search_timer_task = COMPLETED_IDLE_TASK
        self.filters = self.load_filters()
        self.game_search = None
        self.set_service(self.filters.get("service"))
        self.icon_type = self.load_icon_type()
        self.game_store = GameStore(self.service, self.service_media)
        self._game_store_generation = 0
        self.current_view = Gtk.Box()
        self.views = {}

        self.dynamic_categories_game_factories = {
            "recent": self.get_recent_games,
            "missing": self.get_missing_games,
            "running": self.get_running_games,
        }

        for smart_category in categories._SMART_CATEGORIES:
            if smart_category.get_name() not in self.dynamic_categories_game_factories:
                self.dynamic_categories_game_factories[smart_category.get_name()] = smart_category.get_games

        self.accelerators = Gtk.AccelGroup()
        self.add_accel_group(self.accelerators)

        self.connect("delete-event", self.on_window_delete)
        self.connect("configure-event", self.on_window_configure)
        self.connect("realize", self.on_load)
        self.connect("drag-data-received", self.on_drag_data_received)
        self.connect("notify::visible", self.on_visible_changed)
        if self.maximized:
            self.maximize()
        self.init_template()
        self._init_actions()

        # Since system-search-symbolic is already *right there* we'll try to pick some
        # other icon for the button that shows the search popover.
        fallback_filter_icons_names = ["filter-symbolic", "edit-find-replace-symbolic"]
        icon_name = pick_stock_icon(fallback_filter_icons_names, fallback_name="system-search-symbolic")
        filter_button_image: Gtk.Image = self.search_filters_button.get_child()
        filter_button_image.set_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
        self.filter_box_search_name = ""

        # Setup Drag and drop
        self.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        self.drag_dest_add_uri_targets()

        self.set_viewtype_icon(self.current_view_type)

        lutris_icon = Gtk.Image.new_from_icon_name("net.lutris.Lutris", Gtk.IconSize.MENU)
        lutris_icon.set_margin_right(3)

        self.sidebar = LutrisSidebar(self.application)
        self.sidebar.connect("selected-rows-changed", self.on_sidebar_changed)
        # "realize" is order sensitive- must connect after sidebar itself connects the same signal
        self.sidebar.connect("realize", self.on_sidebar_realize)
        self.sidebar_scrolled.add(self.sidebar)

        # This must wait until the selected-rows-changed signal is connected
        self.sidebar.initialize_rows()

        self.sidebar_revealer.set_reveal_child(self.side_panel_visible)
        self.sidebar_revealer.set_transition_duration(300)

        self.game_bar = None
        self.revealer_box = Gtk.HBox(visible=True)
        self.game_revealer.add(self.revealer_box)

        self.update_action_state()
        self.update_notification()

        BUSY_STARTED.register(self.on_busy_started)
        BUSY_STOPPED.register(self.on_busy_stopped)
        SERVICE_LOGIN.register(self.on_service_login)
        SERVICE_LOGOUT.register(self.on_service_logout)
        SERVICE_GAMES_LOADED.register(self.on_service_games_loaded)
        CATEGORIES_UPDATED.register(self.on_categories_updated)
        SAVED_SEARCHES_UPDATED.register(self.on_categories_updated)
        GAME_UPDATED.register(self.on_game_updated)
        GAME_STOPPED.register(self.on_game_stopped)
        GAME_INSTALLED.register(self.on_game_installed)
        GAME_UNHANDLED_ERROR.register(self.on_game_unhandled_error)
        settings.SETTINGS_CHANGED.register(self.on_settings_changed)
        MISSING_GAMES.updated.register(self.update_missing_games_sidebar_row)
        LUTRIS_ACCOUNT_CONNECTED.register(self.on_lutris_account_connected)
        LUTRIS_ACCOUNT_DISCONNECTED.register(self.on_lutris_account_disconnected)
        LOCAL_LIBRARY_UPDATED.register(self.on_local_library_updated)
        THEME_CHANGED.register(self.on_theme_changed)

        # Finally trigger the initialization of the view here
        selected_category = settings.read_setting("selected_category", default="runner:all")
        self.sidebar.selected_category = selected_category.split(":", maxsplit=1) if selected_category else None

        schedule_at_idle(self.sync_library, delay_seconds=1.0)

    def on_busy_started(self):
        display = Gdk.Display.get_default()
        self.get_window().set_cursor(Gdk.Cursor.new_from_name(display, "progress"))

    def on_busy_stopped(self):
        self.get_window().set_cursor(None)

    def _init_actions(self):
        Action = namedtuple("Action", ("callback", "type", "enabled", "default", "accel"))
        Action.__new__.__defaults__ = (None, None, None, None, None)

        actions = {
            "add-game": Action(self.on_add_game_button_clicked),
            "preferences": Action(self.on_preferences_activate),
            "about": Action(self.on_about_clicked),
            "show-installed-only": Action(  # delete?
                self.on_show_installed_state_change,
                type="b",
                default=self.filter_installed,
                accel="<Primary>i",
            ),
            "toggle-viewtype": Action(self.on_toggle_viewtype),
            "toggle-badges": Action(
                self.on_toggle_badges,
                type="b",
                default=settings.read_setting("hide_badges_on_icons"),
                accel="<Primary>p",
            ),
            "icon-type": Action(self.on_icontype_state_change, type="s", default=self.icon_type),
            "view-sorting": Action(
                self.on_view_sorting_state_change,
                type="s",
                default=self.view_sorting,
                enabled=lambda: self.is_view_sort_sensitive,
            ),
            "view-sorting-installed-first": Action(
                self.on_view_sorting_installed_first_change,
                type="b",
                default=self.view_sorting_installed_first,
                enabled=lambda: self.is_view_sort_sensitive,
            ),
            "view-reverse-order": Action(
                self.on_view_sorting_direction_change,
                type="b",
                default=self.view_reverse_order,
                enabled=lambda: self.is_view_sort_sensitive,
            ),
            "show-side-panel": Action(
                self.on_side_panel_state_change,
                type="b",
                default=self.side_panel_visible,
                accel="F9",
            ),
            "show-hidden-games": Action(
                self.on_show_hidden_clicked,
                enabled=lambda: self.is_show_hidden_sensitive,
                accel="<Primary>h",
            ),
            "open-search-filters": Action(self.on_open_search_filters),
            "open-forums": Action(lambda *x: open_uri("https://forums.lutris.net/")),
            "open-discord": Action(lambda *x: open_uri("https://discord.gg/Pnt5CuY")),
            "donate": Action(lambda *x: open_uri("https://lutris.net/donate")),
        }

        self.actions = {}
        self.action_state_updaters = []
        app = self.props.application
        for name, value in actions.items():
            if not value.type:
                action = Gio.SimpleAction.new(name)
                action.connect("activate", value.callback)
            else:
                default_value = None
                param_type = None
                if value.default is not None:
                    default_value = GLib.Variant(value.type, value.default)
                if value.type != "b":
                    param_type = default_value.get_type()
                action = Gio.SimpleAction.new_stateful(name, param_type, default_value)
                action.connect("change-state", value.callback)
            self.actions[name] = action
            if value.enabled:

                def updater(action=action, value=value):
                    action.props.enabled = value.enabled()

                self.action_state_updaters.append(updater)
            self.add_action(action)
            if value.accel:
                app.add_accelerator(value.accel, "win." + name)

    def sync_library(self, force: bool = False) -> None:
        """Tasks that can be run after the UI has been initialized."""

        def on_library_synced(_result, error):
            """Sync media after the library is loaded"""
            if not error:
                AsyncCall(sync_media, None)

        if settings.read_bool_setting("library_sync_enabled", True):
            AsyncCall(LibrarySyncer().sync_local_library, on_library_synced if force else None, force=force)

    def update_action_state(self):
        """This invokes the functions to update the enabled states of all the actions
        which can be disabled."""
        for updater in self.action_state_updaters:
            updater()

    @property
    def service_media(self):
        return self.get_service_media(self.load_icon_type())

    @property
    def selected_category(self):
        return self.sidebar.selected_category

    def on_load(self, widget, data=None):
        """Finish initializing the view"""
        self._bind_zoom_adjustment()
        self.current_view.grab_focus()

    def on_sidebar_realize(self, widget, data=None):
        """Grab the initial focus after the sidebar is initialized - so the view is ready."""
        self.current_view.grab_focus()

    def on_drag_data_received(self, _widget, _drag_context, _x, _y, data, _info, _time):
        """Handler for drop event"""
        file_paths = [unquote(urlparse(uri).path) for uri in data.get_uris()]
        dialog = ImportGameDialog(file_paths, parent=self)
        dialog.show()

    def load_filters(self):
        """Load the initial filters when creating the view"""
        # The main sidebar-category filter will be populated when the sidebar row is selected, after this
        return {"installed": self.filter_installed}

    @property
    def is_show_hidden_sensitive(self) -> bool:
        """True if there are any hidden games to show."""
        return bool(
            self.sidebar.selected_category == ("category", ".hidden")
            or categories_db.get_game_ids_for_categories([".hidden"])
        )

    def on_show_hidden_clicked(self, action, value):
        """Hides or shows the hidden games"""
        hidden_category = "category", ".hidden"
        if self.sidebar.selected_category == hidden_category:
            if self.sidebar.previous_category:
                self.sidebar.selected_category = self.sidebar.previous_category
        else:
            self.sidebar.hidden_row.show()
            self.sidebar.selected_category = hidden_category

    def on_open_search_filters(self, _action, _value):
        def on_filter_popover_closed(_popover):
            self.filter_box_search_name = filter_box.search_name
            self.search_filters_button.set_active(False)

        def on_saved(_box, search_name):
            def switch_to_saved_search():
                self.sidebar.selected_category = "saved_search", search_name
                self.search_entry.set_text("")

            filter_popover.popdown()
            schedule_at_idle(switch_to_saved_search)

        if self.search_filters_button.get_active():
            search = self.get_game_search()
            new_search = saved_searches_db.SavedSearch(0, "", str(search))
            filter_box = SearchFiltersBox(saved_search=new_search, search_entry=self.search_entry)
            filter_box.set_size_request(600, -1)
            if self.filter_box_search_name:
                filter_box.search_name = self.filter_box_search_name
            filter_box.connect("saved", on_saved)
            filter_box.show()
            filter_popover = Gtk.Popover(child=filter_box, can_focus=False, relative_to=self.search_filters_button)
            filter_popover.connect("closed", on_filter_popover_closed)
            filter_popover.popup()

    @property
    def current_view_type(self):
        """Returns which kind of view is currently presented (grid or list)"""
        return settings.read_setting("view_type") or "grid"

    @property
    def filter_installed(self):
        return settings.read_bool_setting("filter_installed", False)

    @property
    def side_panel_visible(self):
        return settings.read_bool_setting("side_panel_visible", True)

    @property
    def show_tray_icon(self):
        """Setting to hide or show status icon"""
        return settings.read_bool_setting("show_tray_icon", False)

    @property
    def view_sorting(self):
        value = settings.read_setting("view_sorting") or "name"
        if value.endswith("_text"):
            value = value[:-5]
        return value

    @property
    def view_reverse_order(self):
        return settings.read_bool_setting("view_reverse_order", False)

    @property
    def view_sorting_installed_first(self):
        return settings.read_bool_setting("view_sorting_installed_first", True)

    @property
    def show_hidden_games(self):
        return settings.read_bool_setting("show_hidden_games", False)

    @property
    def is_view_sort_sensitive(self):
        """True if the view sorting options will be effective; dynamic categories ignore them."""
        return self.filters.get("dynamic_category") not in self.dynamic_categories_game_factories

    def get_sort_sensitive_columns(self) -> Set[int]:
        if self.is_view_sort_sensitive:
            if self.view_sorting == "name":
                return set([COL_NAME, COL_SORTNAME])
            elif self.view_sorting == "year":
                return set([COL_YEAR])
            elif self.view_sorting == "lastplayed":
                return set([COL_LASTPLAYED, COL_LASTPLAYED_TEXT])
            elif self.view_sorting == "installed_at":
                return set([COL_INSTALLED_AT, COL_INSTALLED_AT_TEXT])
            elif self.view_sorting == "playtime":
                return set([COL_PLAYTIME, COL_PLAYTIME_TEXT])

        return set()

    def apply_view_sort(self, items, resolver=lambda i: i):
        """This sorts a list of items according to the view settings of this window;
        the items can be anything, but you can provide a lambda that provides a
        database game dictionary for each one; this dictionary carries the
        data we sort on (though any field may be missing).

        This sort always sorts installed games ahead of uninstalled ones, even when
        the sort is set to descending.

        This treats 'name' sorting specially, applying a natural sort so that
        'Mega slap battler 20' comes after 'Mega slap battler 3'."""
        sort_defaults = {
            "name": "",
            "year": 0,
            "lastplayed": 0.0,
            "installed_at": 0.0,
            "playtime": 0.0,
        }

        def get_sort_default(item):
            """Returns the default value to use when the value is missing; we may be able
            to extract this from the item.."""
            if self.view_sorting == "year" and self.service:
                service_year = self.service.get_game_release_date(item)
                service_year = convert_value(service_year)
                if service_year:
                    return service_year

            # Users may have obsolete view_sorting settings, so
            # we must tolerate them. We treat them all as blank.
            return sort_defaults.get(self.view_sorting, "")

        def convert_value(value):
            """Converts 'value' to the type required for the sort that is in use. Returns None if this
            can't be managed."""
            try:
                if not value:
                    return None
                if self.view_sorting == "name":
                    return str(value)
                if self.view_sorting == "year":
                    # Years can take many forms! We'll try to convert as best we can.
                    if isinstance(value, datetime):
                        return int(value.year)
                    else:
                        try:
                            return int(value)
                        except ValueError:
                            as_date = datetime.strptime(str(value), "%Y-%m-%d")
                            return int(as_date.year)
                else:
                    return float(value)
            except ValueError:
                return None  # unable to parse value?

        def extend_value(value):
            """Expands the value to sort by to a more complex form, for smarter sorting."""
            if self.view_sorting == "name":
                return get_natural_sort_key(value)
            if self.view_sorting == "year":
                contains_year = bool(value)
                if self.view_reverse_order:
                    contains_year = not contains_year
                return contains_year, value
            return value

        def get_sort_value(item):
            db_game = resolver(item)
            if not db_game:
                installation_flag = False
                value = None
            else:
                installation_flag = bool(db_game.get("installed"))

                # When sorting by name, check for a valid sortname first, then fall back
                # on name if valid sortname is not available.
                if self.view_sorting == "name":
                    value = db_game.get("sortname") or db_game.get("name")
                else:
                    value = db_game.get(self.view_sorting)

            value = convert_value(value) or get_sort_default(item)
            value = extend_value(value)

            if self.view_sorting_installed_first:
                # We want installed games to always be first, even in
                # a descending sort.
                if self.view_reverse_order:
                    installation_flag = not installation_flag
                if self.view_sorting == "name":
                    installation_flag = not installation_flag
                return installation_flag, value
            return value

        reverse = self.view_reverse_order if self.view_sorting == "name" else not self.view_reverse_order
        return sorted(items, key=get_sort_value, reverse=reverse)

    def get_running_games(self):
        """Return a list of currently running games"""
        return games_db.get_games_by_ids(self.application.get_running_game_ids())

    def get_missing_games(self):
        games = games_db.get_games_by_ids(MISSING_GAMES.missing_game_ids)
        return self.filter_games(games)

    def update_missing_games_sidebar_row(self) -> None:
        missing_games = self.get_missing_games()
        if missing_games:
            self.sidebar.missing_row.show()
            if self.selected_category == ("dynamic_category", "missing"):
                self.update_store()
        else:
            missing_ids = MISSING_GAMES.missing_game_ids
            if missing_ids:
                logger.warning("Path cache out of date? (%s IDs missing)", len(missing_ids))
            self.sidebar.missing_row.hide()

    def get_recent_games(self):
        """Return a list of currently running games"""
        games = games_db.get_games(filters={"installed": "1"})
        games = self.filter_games(games)
        return sorted(games, key=lambda game: max(game["installed_at"] or 0, game["lastplayed"] or 0), reverse=True)

    def get_game_search(self):
        """Returns a game-search object for the current view settings and search text; this object
        is cached so that we need not re-parse the search if it has not changed."""
        text = self.filters.get("text") or ""
        if self.game_search is None or self.game_search.service != self.service or self.game_search.text != text:
            self.game_search = GameSearch(text, self.service)
        return self.game_search

    def filter_games(self, games, searches: Iterable[GameSearch] = None):
        """Filters a list of games according to the 'installed' and 'text' filters, if those are
        set. But if not, can just return games unchanged."""

        if searches is None:
            search = self.get_game_search()

            if self.filters.get("installed") and not search.has_component("installed"):
                search = search.with_predicate(search.get_installed_predicate(installed=True))

            category = self.filters.get("category") or "all"

            if category != ".hidden" and not search.has_component("hidden"):
                search = search.with_predicate(NotPredicate(search.get_category_predicate(".hidden")))

            searches = [search]

        to_apply = [search for search in searches if not search.is_empty]

        if not to_apply:
            return games

        def matches(game):
            for search in searches:
                if not search.matches(game):
                    return False
            return True

        return [game for game in games if matches(game)]

    def set_service(self, service_name):
        if self.service and self.service.id == service_name:
            return self.service
        if not service_name:
            self.service = None
            return
        try:
            self.service = services.SERVICES[service_name]()
        except KeyError:
            logger.error("Non existent service '%s'", service_name)
            self.service = None
        return self.service

    @staticmethod
    def combine_games(service_game, lutris_game):
        """Inject lutris game information into a service game"""
        if lutris_game and service_game["appid"] == lutris_game["service_id"]:
            for field in ("platform", "runner", "installed_at", "lastplayed", "playtime", "installed"):
                service_game[field] = lutris_game[field]
            service_game["year"] = service_game["year"] if "year" in service_game else lutris_game["year"]
        return service_game

    def get_service_games(self, service_id):
        """Return games for the service indicated."""
        service_games = ServiceGameCollection.get_for_service(service_id)
        for game in service_games:
            game["year"] = self.service.get_game_release_year(game)

        if service_id == "lutris":
            lutris_games = {g["slug"]: g for g in games_db.get_games()}
        else:
            lutris_games = {g["service_id"]: g for g in games_db.get_games(filters={"service": self.service.id})}

        return self.filter_games(
            [
                self.combine_games(game, lutris_games.get(game["appid"]))
                for game in self.apply_view_sort(service_games, lambda game: lutris_games.get(game["appid"]) or game)
            ]
        )

    def get_games_from_filters(self):
        service_id = self.filters.get("service")
        if service_id in services.SERVICES:
            if self.service.online and not self.service.is_authenticated():
                self.show_empty_label()
                return []
            return self.get_service_games(service_id)
        if self.filters.get("dynamic_category") in self.dynamic_categories_game_factories:
            return self.dynamic_categories_game_factories[self.filters["dynamic_category"]]()

        search = self.get_game_search()
        category = self.filters.get("category") or "all"
        searches = [search]

        saved_search = self.filters.get("saved_search")
        if saved_search:
            saved_search_found = saved_searches_db.get_saved_search_by_name(saved_search)

            if saved_search_found:
                try:
                    searches.append(GameSearch(saved_search_found.search, service=None))
                    category = "all"
                except InvalidSearchTermError:
                    pass

        included = [category] if category != "all" else None
        excluded = (
            [".hidden"] if category != ".hidden" and not any(s for s in searches if s.has_component("hidden")) else []
        )
        category_game_ids = categories_db.get_game_ids_for_categories(included, excluded)

        filters = self.get_sql_filters()
        excludes = {}

        if self.filters.get("category") == "all" and "service" not in self.filters:
            excluded_services = set(
                s.casefold()
                for s in services.SERVICES.keys()
                if not settings.read_bool_setting(s + "_in_games_view", default=True, section="services")
            )
            if excluded_services:
                excludes["service"] = excluded_services

        games = games_db.get_games(filters=filters, excludes=excludes)
        games = self.filter_games([game for game in games if game["id"] in category_game_ids], searches=searches)
        return self.apply_view_sort(games)

    def get_sql_filters(self) -> Dict[str, str]:
        """Return the current filters for the view"""
        sql_filters = {}
        if self.filters.get("runner"):
            sql_filters["runner"] = self.filters["runner"]
        if self.filters.get("platform"):
            sql_filters["platform"] = self.filters["platform"]
        if self.filters.get("installed") and not self.get_game_search().has_component("installed"):
            sql_filters["installed"] = "1"

        # We omit the "text" search here because SQLite does a fairly literal
        # search, which is accent sensitive. We'll do better with self.filter_games()
        return sql_filters

    def get_service_media(self, icon_type):
        """Return the ServiceMedia class used for this view"""
        service = self.service if self.service else LutrisService
        medias = service.medias
        if icon_type in medias:
            return medias[icon_type]()
        return medias[service.default_format]()

    def update_revealer(self, games=None):
        if games is not None:  # games can be an empty list!
            if self.game_bar:
                self.game_bar.destroy()
            if len(games) == 1 and games[0]:
                self.game_bar = GameBar(games[0], self.application, self)
                self.revealer_box.pack_start(self.game_bar, True, True, 0)
            else:
                self.game_bar = None
        elif self.game_bar:
            # The game bar can't be destroyed here because the game gets unselected on Wayland
            # whenever the game bar is interacted with. Instead, we keep the current game bar open
            # when the game gets unselected, which is somewhat closer to what the intended behavior
            # should be anyway. Might require closing the game bar manually in some cases.
            pass
            # self.game_bar.destroy()
        if self.revealer_box.get_children():
            self.game_revealer.set_reveal_child(True)
        else:
            self.game_revealer.set_reveal_child(False)

    def show_empty_label(self):
        """Display a label when the view is empty"""
        if self.service and self.service.online and not self.service.is_authenticated():
            self.show_label(_("Connect your %s account to access your games") % self.service.name)
            return

        filter_text = self.filters.get("text")
        has_uninstalled_games = games_db.get_game_count("installed", "0")
        if filter_text:
            if self.filters.get("category") == "favorite":
                self.show_label(_("Add a game matching '%s' to your favorites to see it here.") % filter_text)
            elif self.filters.get("category") == ".hidden":
                self.show_label(_("No hidden games matching '%s' found.") % filter_text)
            elif self.filters.get("installed") and has_uninstalled_games:
                self.show_label(
                    _("No installed games matching '%s' found. Press Ctrl+I to show uninstalled games.") % filter_text
                )
            else:
                self.show_label(_("No games matching '%s' found ") % filter_text)
        else:
            if self.filters.get("category") == "favorite":
                self.show_label(_("Add games to your favorites to see them here."))
            elif self.filters.get("category") == ".hidden":
                self.show_label(_("No games are hidden."))
            elif self.filters.get("installed") and has_uninstalled_games:
                self.show_label(_("No installed games found. Press Ctrl+I to show uninstalled games."))
            elif (
                not self.filters.get("runner")
                and not self.filters.get("service")
                and not self.filters.get("platform")
                and not self.filters.get("dynamic_category")
            ):
                self.show_splash()
            else:
                self.show_label(_("No games found"))

    def refresh_view(self):
        self.sidebar.update_rows()
        self.update_missing_games_sidebar_row()
        self.update_store()

    def update_store(self) -> None:
        service_id = self.filters.get("service")
        service = self.service
        service_media = self.service_media
        self._game_store_generation += 1
        generation = self._game_store_generation

        def make_game_store(games):
            game_store = GameStore(service, service_media)
            game_store.add_preloaded_games(games, service_id)
            return games, game_store

        def on_games_ready(games, error):
            if generation != self._game_store_generation:
                return  # no longer applicable, we got switched again!

            if error:
                raise error  # bounce any error against the backstop

            # Since get_games_from_filters() seems to be much faster than making a GameStore,
            # we defer the spinner to here, when we know how many games we will show. If there
            # are "many" we show a spinner while the store is built.
            if len(games) > 512:
                self.show_spinner()

            AsyncCall(make_game_store, apply_store, games)

        def apply_store(result, error):
            if generation != self._game_store_generation:
                return  # no longer applicable, we got switched again!

            if error:
                raise error  # bounce any error against the backstop

            games, game_store = result

            placeholder_text = self._get_search_placeholder_text(games)
            self.search_entry.set_placeholder_text(placeholder_text)

            for view in self.views.values():
                view.service = self.service

            GLib.idle_add(self.update_revealer)

            if self.game_store != game_store:
                self.game_store = game_store

                view_type = self.current_view_type

                if view_type in self.views:
                    view = self.views[view_type]
                    self.current_view = view
                    selected_ids = [view.get_game_id_for_path(p) for p in view.get_selected()]
                    view.set_game_store(self.game_store)
                    new_selection = [view.get_path_for_game_id(game_id) for game_id in selected_ids]
                    view.set_selected(filter(None, new_selection), scroll_into_view=True)

            if games:
                self.hide_overlay()
            else:
                self.show_empty_label()

            self.update_notification()

        AsyncCall(self.get_games_from_filters, on_games_ready)

    @staticmethod
    def _get_search_placeholder_text(games) -> str:
        if not games:
            return _("Search games")

        games_count = len(games)
        return ngettext("Search %d game", "Search %d games", games_count) % games_count

    def _bind_zoom_adjustment(self):
        """Bind the zoom slider to the supported banner sizes"""
        service = self.service if self.service else LutrisService
        icon_type = self.load_icon_type()
        self.zoom_adjustment.set_lower(0)
        self.zoom_adjustment.set_upper(len(service.medias) - 1)

        if icon_type not in service.medias:
            icon_type = service.default_format

        try:
            value = list(service.medias.keys()).index(icon_type)
        except ValueError:
            value = 0

        self.zoom_adjustment.props.value = value
        self.zoom_adjustment.connect("value-changed", self.on_zoom_changed)

    def on_zoom_changed(self, adjustment):
        """Handler for zoom modification"""
        media_index = round(adjustment.props.value)
        adjustment.props.value = media_index
        service = self.service if self.service else LutrisService
        media_services = list(service.medias.keys())
        if len(media_services) <= media_index:
            media_index = media_services.index(service.default_format)
        icon_type = media_services[media_index]
        if icon_type != self.icon_type:
            GLib.idle_add(self.save_icon_type, icon_type)

    def show_label(self, message):
        """Display a label in the middle of the UI"""
        self.show_overlay(Gtk.Label(message, visible=True))

    def show_splash(self):
        theme = "dark" if self.application.style_manager.is_dark else "light"
        side_splash = Gtk.Image(visible=True)
        side_splash.set_from_file(os.path.join(datapath.get(), "media/side-%s.svg" % theme))
        side_splash.set_alignment(0, 0)

        center_splash = Gtk.Image(visible=True)
        center_splash.set_alignment(0.5, 0.5)
        center_splash.set_from_file(os.path.join(datapath.get(), "media/splash-%s.svg" % theme))

        splash_box = Gtk.HBox(visible=True, margin_top=24)
        splash_box.pack_start(side_splash, False, False, 12)
        splash_box.set_center_widget(center_splash)
        splash_box.is_splash = True
        self.show_overlay(splash_box, Gtk.Align.FILL, Gtk.Align.FILL)

    def is_showing_splash(self):
        if self.blank_overlay.get_visible():
            for ch in self.blank_overlay.get_children():
                if hasattr(ch, "is_splash"):
                    return True
        return False

    def on_theme_changed(self):
        if self.is_showing_splash():
            self.show_splash()

    def show_spinner(self):
        # This is inconsistent, but we can't use the blank overlay for the spinner- it
        # won't reliably start as a child of blank_overlay. It seems like it fails if
        # blank_overlay has never yet been visible.
        # It works better if created up front and shown like this.
        self.game_view_spinner.start()
        self.game_view_spinner.show()
        self.games_stack.hide()
        self.blank_overlay.hide()

    def show_overlay(self, widget, halign=Gtk.Align.FILL, valign=Gtk.Align.FILL):
        """Display a widget in the blank overlay"""
        for child in self.blank_overlay.get_children():
            child.destroy()
        self.blank_overlay.set_halign(halign)
        self.blank_overlay.set_valign(valign)
        self.blank_overlay.add(widget)
        self.blank_overlay.show()
        self.games_stack.hide()
        self.game_view_spinner.hide()

    def hide_overlay(self):
        self.blank_overlay.hide()
        self.game_view_spinner.hide()
        self.games_stack.show()
        for child in self.blank_overlay.get_children():
            child.destroy()

    @property
    def view_type(self):
        """Return the type of view saved by the user"""
        view_type = settings.read_setting("view_type")
        if view_type in ["grid", "list"]:
            return view_type
        return self.default_view_type

    def do_key_press_event(self, event):  # pylint: disable=arguments-differ
        # XXX: This block of code below is to enable searching on type.
        # Enabling this feature steals focus from other entries so it needs
        # some kind of focus detection before enabling library search.

        # Probably not ideal for non-english, but we want to limit
        # which keys actually start searching
        if event.keyval == Gdk.KEY_Escape:
            self.search_entry.set_text("")
            self.current_view.grab_focus()
            return Gtk.ApplicationWindow.do_key_press_event(self, event)

        if (  # pylint: disable=too-many-boolean-expressions
            not Gdk.KEY_0 <= event.keyval <= Gdk.KEY_z
            or event.state & Gdk.ModifierType.CONTROL_MASK
            or event.state & Gdk.ModifierType.SHIFT_MASK
            or event.state & Gdk.ModifierType.META_MASK
            or event.state & Gdk.ModifierType.MOD1_MASK
            or self.search_entry.has_focus()
        ):
            return Gtk.ApplicationWindow.do_key_press_event(self, event)
        self.search_entry.grab_focus()
        return self.search_entry.do_key_press_event(self.search_entry, event)

    def load_icon_type(self):
        """Return the icon style depending on the type of view."""
        default_icon_types = {
            "icon_type_grid": "coverart_med",
        }
        setting_key = "icon_type_%sview" % self.current_view_type
        if self.service and self.service.id != "lutris":
            setting_key += "_%s" % self.service.id
        self.icon_type = settings.read_setting(setting_key, default=default_icon_types.get(setting_key, ""))
        return self.icon_type

    def save_icon_type(self, icon_type):
        """Save icon type to settings"""
        self.icon_type = icon_type
        setting_key = "icon_type_%sview" % self.current_view_type
        if self.service and self.service.id != "lutris":
            setting_key += "_%s" % self.service.id
        settings.write_setting(setting_key, self.icon_type)
        self.redraw_view()

    def redraw_view(self):
        """Completely reconstruct the main view"""
        if not self.game_store:
            logger.error("No game store yet")
            return

        view_type = self.current_view_type

        if view_type not in self.views:
            self.game_store = GameStore(self.service, self.service_media)
            if view_type == "grid":
                self.current_view = GameGridView(
                    self.game_store, hide_text=settings.read_bool_setting("hide_text_under_icons")
                )
            else:
                self.current_view = GameListView(self.game_store)

            self.current_view.connect("game-selected", self.on_game_selection_changed)
            self.current_view.connect("game-activated", self.on_game_activated)
            self.views[view_type] = self.current_view

        scrolledwindow = self.games_stack.get_child_by_name(view_type)

        if not scrolledwindow:
            scrolledwindow = Gtk.ScrolledWindow()
            self.games_stack.add_named(scrolledwindow, view_type)

        if not scrolledwindow.get_child():
            scrolledwindow.add(self.current_view)
            scrolledwindow.show_all()

        self.update_view_settings()
        self.games_stack.set_visible_child_name(view_type)
        self.update_action_state()
        self.update_store()

    def rebuild_view(self, view_type):
        """Discards the view named by 'view_type' and if it is the current view,
        regenerates it. This is used to update view settings that can only be
        set during view construction, and not updated later."""
        if view_type in self.views:
            view = self.views[view_type]
            scrolledwindow = self.games_stack.get_child_by_name(view_type)
            scrolledwindow.remove(view)
            del self.views[view_type]
            if self.current_view_type == view_type:
                self.redraw_view()
            # Because the view has hooks and such hooked up, it must be explicitly
            # destroyed to disconnect everything.
            view.destroy()

    def update_view_settings(self):
        if self.current_view and self.current_view_type == "grid":
            show_badges = settings.read_setting("hide_badges_on_icons") != "True"
            self.current_view.show_badges = show_badges and not bool(self.filters.get("platform"))

    def set_viewtype_icon(self, view_type):
        self.viewtype_icon.set_from_icon_name("view-%s-symbolic" % view_type, Gtk.IconSize.BUTTON)

    def set_show_installed_state(self, filter_installed):
        """Shows or hide uninstalled games"""
        settings.write_setting("filter_installed", bool(filter_installed))
        self.filters["installed"] = filter_installed

    def update_notification(self):
        show_notification = self.is_showing_splash()
        if show_notification:
            self.lutris_log_in_label.show()
        self.login_notification_revealer.set_reveal_child(show_notification)

    @GtkTemplate.Callback
    def on_lutris_log_in_label_activate_link(self, _label, _url):
        def on_connect_success(widget, _username):
            self.sync_library(force=True)

        self.login_notification_revealer.set_reveal_child(False)
        login_dialog = ClientLoginDialog(parent=self)
        login_dialog.connect("connected", on_connect_success)

    def on_version_notification_close_button_clicked(self, _button):
        dialog = QuestionDialog(
            {
                "title": _("Unsupported Lutris Version"),
                "question": _(
                    "This version of Lutris will no longer receive support on Github and Discord, "
                    "and may not interoperate properly with Lutris.net. Do you want to use it anyway?"
                ),
                "parent": self,
            }
        )

        if dialog.result == Gtk.ResponseType.YES:
            self.version_notification_revealer.set_reveal_child(False)
            runtime_versions = get_runtime_versions()
            if runtime_versions:
                client_version = runtime_versions.get("client_version")
                settings.write_setting("ignored_supported_lutris_verison", client_version or "")

    def on_service_games_loaded(self, service):
        """Request a view update when service games are loaded"""
        if self.service and service.id == self.service.id:
            self.update_store()

    def on_categories_updated(self):
        self.update_store()

    def save_window_state(self):
        """Saves the window's size position and state as settings."""
        width, height = self.window_size
        settings.write_setting("width", width)
        settings.write_setting("height", height)
        if self.window_x and self.window_y:
            settings.write_setting("window_x", self.window_x)
            settings.write_setting("window_y", self.window_y)
        settings.write_setting("maximized", self.maximized)

    def restore_window_position(self):
        """Restores the window position only; we call this when showing
        the window, but restore the other settings only when creating it."""
        self.window_x = settings.read_setting("window_x")
        self.window_y = settings.read_setting("window_y")
        if self.window_x and self.window_y:
            self.move(int(self.window_x), int(self.window_y))

    def on_service_login(self, service):
        self.update_notification()
        service.start_reload(self._service_reloaded_cb)
        return True

    def _service_reloaded_cb(self, error):
        if error:
            dialogs.display_error(error, parent=self)

    def on_service_logout(self, service):
        self.update_notification()
        if self.service and service.id == self.service.id:
            self.update_store()
        return True

    def on_lutris_account_connected(self):
        self.update_notification()
        self.sync_library(force=True)

    def on_lutris_account_disconnected(self):
        self.update_notification()

    def on_local_library_updated(self):
        self.redraw_view()

    @GtkTemplate.Callback
    def on_resize(self, widget, *_args):
        """Size-allocate signal.
        Updates stored window size and maximized state.
        """
        if not widget.get_window():
            return
        self.maximized = widget.is_maximized()
        size = widget.get_size()
        if not self.maximized:
            self.window_size = size
        self.search_entry.set_size_request(min(max(50, size[0] - 470), 800), -1)

    def on_window_delete(self, *_args):
        app = self.application
        if app.has_running_games:
            self.hide()
            return True
        if app.has_tray_icon():
            self.hide()
            return True

    def on_visible_changed(self, window, param):
        if self.application.tray:
            self.application.tray.update_present_menu()

    def on_window_configure(self, *_args):
        """Callback triggered when the window is moved, resized..."""
        self.window_x, self.window_y = self.get_position()

    @GtkTemplate.Callback
    def on_destroy(self, *_args):
        """Signal for window close."""
        # Stop cancellable running threads
        for stopper in self.threads_stoppers:
            stopper()

    @GtkTemplate.Callback
    def on_hide(self, *_args):
        self.save_window_state()

    @GtkTemplate.Callback
    def on_show(self, *_args):
        self.restore_window_position()

    @GtkTemplate.Callback
    def on_preferences_activate(self, *_args):
        """Callback when preferences is activated."""
        self.application.show_window(PreferencesDialog, parent=self)

    def on_show_installed_state_change(self, action, value):
        """Callback to handle uninstalled game filter switch"""
        action.set_state(value)
        self.set_show_installed_state(value.get_boolean())
        self.update_store()

    @GtkTemplate.Callback
    def on_search_entry_changed(self, entry):
        """Callback for the search input keypresses"""
        self.search_timer_task.unschedule()
        self.filters["text"] = entry.get_text().strip()
        self.search_timer_task = schedule_at_idle(self.update_store, delay_seconds=0.5)

    @GtkTemplate.Callback
    def on_search_entry_key_press(self, widget, event):
        if event.keyval == Gdk.KEY_Down:
            if self.current_view_type == "grid":
                self.current_view.select_path(Gtk.TreePath("0"))  # needed for gridview only
                # if game_bar is alive at this point it can mess grid item selection up
                # for some unknown reason,
                # it is safe to close it here, it will be reopened automatically.
                if self.game_bar:
                    self.game_bar.destroy()  # for gridview only
            self.current_view.set_cursor(Gtk.TreePath("0"), None, False)  # needed for both view types
            self.current_view.grab_focus()

    @GtkTemplate.Callback
    def on_about_clicked(self, *_args):
        """Open the about dialog."""
        dialogs.AboutDialog(parent=self)

    def on_game_unhandled_error(self, _game: Game, error: BaseException) -> None:
        """Called when a game has sent the 'game-error' signal"""

        error_handler = get_error_handler(type(error))
        error_handler(error, self)

    @GtkTemplate.Callback
    def on_add_game_button_clicked(self, *_args):
        """Add a new game manually with the AddGameDialog."""
        self.application.show_window(AddGamesWindow, parent=self)
        return True

    def on_toggle_viewtype(self, *args):
        view_type = "list" if self.current_view_type == "grid" else "grid"
        logger.debug("View type changed to %s", view_type)
        self.set_viewtype_icon(view_type)
        settings.write_setting("view_type", view_type)
        self.redraw_view()
        self._bind_zoom_adjustment()

    def on_icontype_state_change(self, action, value):
        action.set_state(value)
        self._set_icon_type(value.get_string())

    def on_view_sorting_state_change(self, action, value):
        self.actions["view-sorting"].set_state(value)
        value = str(value).strip("'")
        settings.write_setting("view_sorting", value)
        self.update_store()

    def on_view_sorting_direction_change(self, action, value):
        self.actions["view-reverse-order"].set_state(value)
        settings.write_setting("view_reverse_order", bool(value))
        self.update_store()

    def on_view_sorting_installed_first_change(self, action, value):
        self.actions["view-sorting-installed-first"].set_state(value)
        settings.write_setting("view_sorting_installed_first", bool(value))
        self.update_store()

    def on_side_panel_state_change(self, action, value):
        """Callback to handle side panel toggle"""
        action.set_state(value)
        side_panel_visible = value.get_boolean()
        settings.write_setting("side_panel_visible", bool(side_panel_visible))
        self.sidebar_revealer.set_reveal_child(side_panel_visible)

    def on_sidebar_changed(self, widget):
        """Handler called when the selected element of the sidebar changes"""
        for filter_type in ("category", "dynamic_category", "saved_search", "service", "runner", "platform"):
            if filter_type in self.filters:
                self.filters.pop(filter_type)

        row_type, row_id = widget.selected_category
        if row_type == "user_category":
            row_type = "category"
        self.filters[row_type] = row_id

        service_name = self.filters.get("service")
        self.set_service(service_name)
        self._bind_zoom_adjustment()
        self.redraw_view()

        if row_type != "category" or row_id != ".hidden":
            self.sidebar.hidden_row.hide()
            self.show_hidden_games_button.set_label(_("Show Hidden Games"))
        else:
            self.show_hidden_games_button.set_label(_("Rehide Hidden Games"))
        # We just _replaced_ the label, need to align it. That is weird and
        # contrary to the docs, but here we are.
        self.show_hidden_games_button.get_child().set_halign(Gtk.Align.START)

        if not MISSING_GAMES.is_initialized or (row_type == "dynamic_category" and row_id == "missing"):
            MISSING_GAMES.update_all_missing()

    def on_game_selection_changed(self, view, selection):
        game_ids = [view.get_game_id_for_path(path) for path in selection]

        games = []
        for game_id in game_ids:
            if self.service:
                game = ServiceGameCollection.get_game(self.service.id, game_id)
            else:
                game = games_db.get_game_by_field(game_id, "id")

            # There can be no game found if you are removing a game; it will
            # still have a selected icon in the UI just long enough to get here.
            if game:
                games.append(game)

        GLib.idle_add(self.update_revealer, games)
        return False

    def on_toggle_badges(self, _widget, _data):
        """Event handler to toggle badge visibility"""
        state = settings.read_setting("hide_badges_on_icons").lower() == "true"
        settings.write_setting("hide_badges_on_icons", not state)
        self.on_settings_changed(None, not state, "hide_badges_on_icons")

    def on_settings_changed(self, setting_key, new_value, section):
        if section == "lutris" and setting_key == "hide_text_under_icons":
            self.rebuild_view("grid")
        elif section == "services" and setting_key.endswith("_in_games_view"):
            self.update_store()
        else:
            self.update_view_settings()
        self.update_notification()
        return True

    def is_game_displayed(self, game):
        """Return whether a game should be displayed on the view"""
        row = self.sidebar.get_selected_row()

        if row:
            # Stopped games do not get displayed on the running page
            if row.type == "dynamic_category" and row.id == "running" and game.state == game.STATE_STOPPED:
                return False

            # If the update took the row out of this view's category, we'll need
            # to update the view to reflect that.
            search = self.get_game_search()
            enforce_hidden = not search.has_component("hidden")
            if row.type == "dynamic_category" and row.id in ("recent", "missing"):
                if enforce_hidden and ".hidden" in game.get_categories():
                    return False
            elif row.type in ("category", "user_category", "saved_search"):
                categories = game.get_categories()
                if enforce_hidden and row.id != ".hidden" and ".hidden" in categories:
                    return False

                if row.id != "all" and row.id not in categories:
                    return False

        return True

    def on_game_updated(self, game):
        """Updates an individual entry in the view when a game is updated"""
        add_to_path_cache(game)
        self.update_action_state()

        if self.service:
            db_game = self.service.get_service_db_game(game)
        else:
            db_game = games_db.get_game_by_field(game.id, "id")

            if db_game and not self.is_game_displayed(game) and "id" in db_game:
                self.game_store.remove_game(db_game["id"])
                return True

        if db_game:
            updated_columns = self.game_store.update(db_game)
            sensitive_columns = self.get_sort_sensitive_columns()
            if updated_columns is None or not sensitive_columns.isdisjoint(updated_columns):
                self.update_store()

        return True

    def on_game_stopped(self, game: Game) -> None:
        """Updates the game list when a game stops; this keeps the 'running' page updated."""
        selected_row = self.sidebar.get_selected_row()
        # Only update the running page- we lose the selected row when we do this,
        # but on the running page this is okay.
        if isinstance(selected_row, SidebarRow) and selected_row.id == "running":
            self.game_store.remove_game(game.id)

    def on_game_installed(self, game):
        self.sync_library()

    def on_game_removed(self):
        """Simple method used to refresh the view"""
        self.refresh_view()
        return True

    def on_game_activated(self, _view, game_id):
        """Handles view activations (double click, enter press)"""
        if self.service:
            logger.debug("Looking up %s game %s", self.service.id, game_id)
            db_game = games_db.get_game_for_service(self.service.id, game_id)

            if db_game and db_game["installed"]:
                game_id = db_game["id"]
            else:
                game_id = self.service.install_by_id(game_id)

        if game_id:
            game = Game(game_id)
            if game.is_installed:
                game.launch(launch_ui_delegate=self)
            else:
                game.install(launch_ui_delegate=self)

    @property
    def download_queue(self) -> DownloadQueue:
        queue = cast(DownloadQueue, self.download_revealer.get_child())
        if not queue:
            queue = DownloadQueue(self.download_revealer)
            self.download_revealer.add(queue)
        return queue

    def start_runtime_updates(self, force_updates: bool) -> None:
        """Starts the process of applying runtime updates, asynchronously. No UI appears until
        we can determine that there are updates to perform."""

        def create_runtime_updater():
            """This function runs on a worker thread and decides what component updates are
            required; we do this on a thread because it involves hitting the Lutris.net website,
            which can easily block."""
            runtime_updater = RuntimeUpdater(force=force_updates)
            component_updaters = runtime_updater.create_component_updaters()
            supported_client_version = runtime_updater.check_client_versions()
            return component_updaters, runtime_updater, supported_client_version

        def create_runtime_updater_cb(result, error):
            """Picks up the component updates when we know what they are, and begins the installation.
            This must be done on the main thread, since it updates the UI. This would be so much less
            ugly with asyncio, but here we are."""
            if error:
                logger.exception("Failed to obtain updates from Lutris.net: %s", error)
            else:
                component_updaters, runtime_updater, supported_client_version = result

                if supported_client_version and not LINUX_SYSTEM.is_flatpak():
                    markup = self.version_notification_label.get_label()
                    markup = markup % (settings.VERSION, supported_client_version)
                    self.version_notification_label.set_label(markup)
                    self.version_notification_revealer.set_reveal_child(True)

                if component_updaters:
                    self.install_runtime_component_updates(component_updaters, runtime_updater)
                else:
                    logger.debug("Runtime up to date")

        AsyncCall(create_runtime_updater, create_runtime_updater_cb)

    def install_runtime_component_updates(
        self,
        updaters: List[ComponentUpdater],
        runtime_updater: RuntimeUpdater,
        completion_function: DownloadQueue.CompletionFunction = None,
        error_function: DownloadQueue.ErrorFunction = None,
    ) -> bool:
        """Installs a list of component updates. This displays progress bars
        in the sidebar as it installs updates, one at a time."""

        queue = self.download_queue
        operation_names = [f"component_update:{u.name}" for u in updaters]

        def install_updates():
            for updater in updaters:
                updater.install_update(runtime_updater)
            for updater in updaters:
                updater.join()

            # better safe than sorry - there are Proton builds outside our control
            clear_wine_version_cache()

        return queue.start_multiple(
            install_updates,
            (u.get_progress for u in updaters),
            completion_function=completion_function,
            error_function=error_function,
            operation_names=operation_names,
        )


def _handle_esynclimiterror(error: EsyncLimitError, parent: Gtk.Window) -> None:
    message = _(
        "Your limits are not set correctly."
        " Please increase them as described here:"
        " <a href='https://github.com/lutris/docs/blob/master/HowToEsync.md'>"
        "How-to:-Esync (https://github.com/lutris/docs/blob/master/HowToEsync.md)</a>"
    )
    ErrorDialog(error, message_markup=message, parent=parent)


register_error_handler(EsyncLimitError, _handle_esynclimiterror)



import json
import os
import platform
from datetime import datetime

try:
    from webdav4.client import Client

    WEBDAV_AVAILABLE = True
except ImportError:
    WEBDAV_AVAILABLE = False

from lutris import settings
from lutris.game import Game
from lutris.util.log import logger
from lutris.util.strings import human_size
from lutris.util.wine.prefix import find_prefix

DIR_CREATE_CACHE = []
SAVE_TYPES = ["saves", "logs", "config", "screenshots"]
SYNC_TYPES = ["saves", "config"]


class SaveInfo:
    save_types = ["saves", "logs", "configs", "screenshots"]
    default_synced_types = ["saves", "configs"]

    def __init__(self, game: Game):
        self.game = game
        if not self.game.config:
            raise ValueError("%s has no configuration" % self.game)
        save_config = self.game.config.game_level["game"].get("save_config")
        if not save_config:
            raise ValueError("%s has no save configuration" % self.game)
        self.save_config: dict = save_config
        self.basedir = self.get_basedir()

    def get_basedir(self) -> str:
        if not self.game.config:
            raise ValueError("%s has no configuration" % self.game)
        save_config = self.game.config.game_level["game"]["save_config"]
        basedir = save_config.get("basedir") or self.game.directory
        if not basedir:
            raise ValueError("No save directory provided")
        prefix_path = os.path.dirname(str(self.game.config.game_config.get("exe")))

        if self.game.runner_name == "wine":
            prefix_path = self.game.config.game_config.get("prefix", "")
            if prefix_path and not prefix_path.startswith("/"):
                prefix_path = os.path.join(self.game.directory, prefix_path)
            exe = self.game.config.game_config.get("exe", "")
            if exe and not exe.startswith("/"):
                exe = os.path.join(self.game.directory, exe)
            if not prefix_path:
                prefix_path = find_prefix(exe)
            if "$GAMEDIR" in basedir:
                basedir = basedir.replace("$GAMEDIR", self.game.directory or prefix_path)
            if "$USERDIR" in basedir:
                username = os.getenv("USER") or "steamuser"
                basedir = basedir.replace("$USERDIR", os.path.join(prefix_path, "drive_c/users/%s" % username))
        else:
            if "$GAMEDIR" in basedir:
                basedir = basedir.replace("$GAMEDIR", self.game.directory)
            if "$USERDIR" in basedir:
                basedir = os.path.expanduser(basedir.replace("$USERDIR", "~"))
        return basedir

    @staticmethod
    def get_dir_info(path: str) -> dict:
        path_files = {}
        path = path.rstrip("/")
        if os.path.isfile(path):
            path_files[os.path.basename(path)] = os.stat(path)
            return path_files
        for root, _dirs, files in os.walk(path):
            basedir = root[len(path) + 1 :]
            for path_file in files:
                path_files[os.path.join(basedir, path_file)] = os.stat(os.path.join(root, path_file))
        return path_files

    def format_dir_info(self, path: str) -> list:
        path_files = self.get_dir_info(path)
        output = []
        for path_file in sorted(path_files, key=lambda k: path_files[k].st_mtime, reverse=True):
            fstats = path_files[path_file]
            output.append(
                {
                    "file": path_file,
                    "size": fstats.st_size,
                    "modified": fstats.st_mtime,
                }
            )
        return output

    def get_save_files(self) -> dict:
        """Return list of files with their details for each type along with metadata related to the save info"""
        results: dict = {"basedir": self.basedir, "config": self.save_config}
        for section in self.save_types:
            if section in self.save_config:
                section_data: dict = {}
                path = os.path.join(self.basedir, self.save_config[section])
                section_data["path"] = path
                section_data["files"] = self.format_dir_info(path)
                results[section] = section_data
        return results

    def print_dir_details(self, title: str, path: str):
        save_files = self.get_dir_info(path)
        if not save_files:
            return
        print(title)
        print("=" * len(title))
        total_size = 0
        for save_file in sorted(save_files, key=lambda k: save_files[k].st_mtime, reverse=True):
            fstats = save_files[save_file]
            total_size += fstats.st_size
            print(
                "%s (%s)\t\t%s"
                % (save_file, human_size(fstats.st_size), datetime.fromtimestamp(fstats.st_mtime).strftime("%c"))
            )
        print("Total size: %s" % human_size(total_size))


def show_save_stats(game, output_format="text"):
    try:
        save_info = SaveInfo(game)
    except ValueError:
        logger.error("Save information for %s unavailable", game)
        return

    if output_format == "json":
        print(json.dumps(save_info.get_save_files()), indent=2)
    else:
        for section in save_info.save_types:
            if section in save_info.save_config:
                save_info.print_dir_details(
                    section.capitalize(), os.path.join(save_info.basedir, save_info.save_config[section])
                )


def create_dirs(client, path):
    parts = path.split("/")
    for i in range(len(parts)):
        relpath = os.path.join(*parts[: i + 1])
        if relpath in DIR_CREATE_CACHE:
            continue
        if not client.exists(relpath):
            logger.debug("Creating Webdav folder %s", relpath)
            client.mkdir(relpath)
        DIR_CREATE_CACHE.append(relpath)


def get_webdav_client():
    if not WEBDAV_AVAILABLE:
        logger.error("Python package 'webdav4' not installed.")
        return
    webdav_host = settings.read_setting("webdav_host")
    if not webdav_host:
        logger.error("No remote host set (webdav_host)")
        return
    webdav_user = settings.read_setting("webdav_user")
    if not webdav_user:
        logger.error("No remote username set (webdav_user)")
        return
    webdav_pass = settings.read_setting("webdav_pass")
    if not webdav_pass:
        logger.error("No remote password set (webdav_pass)")
        return
    return Client(webdav_host, auth=(webdav_user, webdav_pass), timeout=50)


def get_existing_saves(client, game_base_dir):
    if not client.exists(game_base_dir):
        return []
    base_dir_content = client.ls(game_base_dir)
    saves = []
    for save_folder in base_dir_content:
        if save_folder["type"] != "directory":
            continue
        local_save_info = os.path.join(settings.CACHE_DIR, "%s.json" % os.path.basename(save_folder["name"]))
        client.download_file(os.path.join(save_folder["name"], "saveinfo.json"), local_save_info)
        saves.append(local_save_info)
    return saves


def upload_save(game, sections=None):
    if not sections:
        sections = SYNC_TYPES
    try:
        save_info = SaveInfo(game)
    except ValueError:
        logger.error("Can't get save info for %s", game)
        return

    print("Uploading save for %s" % game)
    webdav_saves_path = settings.read_setting("webdav_saves_path")
    if not webdav_saves_path:
        logger.error("No save path for the remote host (webdav_saves_path setting)")
        return
    client = get_webdav_client()
    if not client:
        return

    # game_base_dir = os.path.join(webdav_saves_path, game.slug)
    # existing_saves = get_existing_saves(client, game_base_dir)
    # for save in existing_saves:
    #     print(save)
    #     os.remove(save)
    save_files = save_info.get_save_files()
    max_time = 0
    for section in sections:
        for save_file in save_files[section].get("files", []):
            if int(save_file["modified"]) > max_time:
                max_time = int(save_file["modified"])

    save_id = f"{platform.node()}-{max_time}"
    save_dest_dir = os.path.join(webdav_saves_path, game.slug, save_id)
    create_dirs(client, save_dest_dir)

    save_info_name = "saveinfo.json"
    save_info_path = os.path.join(settings.CACHE_DIR, save_info_name)
    with open(save_info_path, "w", encoding="utf-8") as save_info_file:
        json.dump(save_files, save_info_file, indent=2)
    client.upload_file(save_info_path, os.path.join(save_dest_dir, save_info_name))
    basepath = save_files["saves"]["path"]
    if os.path.isfile(basepath):
        basepath = os.path.dirname(basepath)
    relpath = basepath[len(save_files["basedir"]) + 1 :]
    create_dirs(client, os.path.join(save_dest_dir, relpath))
    for save_file in save_files["saves"]["files"]:
        upload_file_source = os.path.join(basepath, save_file["file"])
        upload_file_dest = os.path.join(save_dest_dir, relpath, save_file["file"])
        print(upload_file_source, ">", upload_file_dest)
        client.upload_file(upload_file_source, upload_file_dest)


def load_save_info(save_info_path):
    with open(save_info_path, "r", encoding="utf-8") as save_info_file:
        save_info = json.load(save_info_file)
    return save_info


def parse_save_info(save_info_path: str):
    """Parse the filename of a save info file and extract its information"""
    save_info_name, _ext = os.path.splitext(os.path.basename(save_info_path))
    hostname, ts = save_info_name.rsplit("-", maxsplit=1)
    return {"hostname": hostname, "datetime": datetime.fromtimestamp(int(ts))}


def save_check(game):
    try:
        save_info = SaveInfo(game)
    except ValueError:
        logger.error("%s has no save configuration", game)
        return

    print("Checking sync of save for %s" % game)
    webdav_saves_path = settings.read_setting("webdav_saves_path")
    if not webdav_saves_path:
        logger.error("No save path for the remote host (webdav_saves_path setting)")
        return
    client = get_webdav_client()
    if not client:
        return

    remote_basedir = os.path.join(webdav_saves_path, game.slug)
    existing_saves = get_existing_saves(client, remote_basedir)
    current_save_files = save_info.get_save_files()
    for save_path in existing_saves:
        save_info_meta = parse_save_info(save_path)
        host = save_info_meta["hostname"]
        print("Host: %s (%s)" % (host, save_info_meta["datetime"].strftime("%c")))
        remote_save_info = load_save_info(save_path)
        unsynced = {}
        for section in SAVE_TYPES:
            if section not in current_save_files:
                continue
            files = {file_info["file"]: file_info for file_info in remote_save_info[section].get("files", [])}
            local_files = {file_info["file"]: file_info for file_info in current_save_files[section].get("files", [])}
            unsynced[section] = {"unsynced": [], "newer": [], "older": [], "missing": []}

            for filename, file_info in local_files.items():
                remote_file = files.get(filename)
                if not remote_file:
                    unsynced[section]["unsynced"].append(filename)
                files[filename]["seen"] = True
                if file_info["modified"] > files[filename]["modified"]:
                    print("Local file %s is newer" % filename)
                    unsynced[section]["newer"].append(filename)
                if file_info["modified"] < files[filename]["modified"]:
                    print("Local file %s is older" % filename)
                    unsynced[section]["older"].append(filename)
            for filename, file_info in files.items():
                if file_info.get("seen"):
                    continue
                unsynced[section]["missing"].append(filename)

        out_of_sync = False
        for section in SAVE_TYPES:
            if section not in unsynced:
                continue
            for _key, value in unsynced[section].items():
                if value:
                    out_of_sync = True
        if not out_of_sync:
            print("🟢 Save %s synced with local game" % os.path.basename(save_path))
        else:
            print("🟠 Save %s out of sync with local game" % os.path.basename(save_path))
            print(unsynced)

        os.remove(save_path)



# Standard Library
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class openmsx(Runner):
    human_name = _("openMSX")
    description = _("MSX computer emulator")
    platforms = [_("MSX, MSX2, MSX2+, MSX turboR")]
    flatpak_id = "org.openmsx.openMSX"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "help": _("The game data, commonly called a ROM image."),
        }
    ]

    def play(self):
        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        return {"command": self.get_command() + [rom]}



import os
from gettext import gettext as _

from lutris.runners.runner import Runner
from lutris.util import display
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.strings import split_arguments


class zdoom(Runner):
    # http://zdoom.org/wiki/Command_line_parameters
    description = _("GZDoom Game Engine")
    human_name = _("GZDoom")
    platforms = [_("Linux")]
    runner_executable = "zdoom/gzdoom"
    flatpak_id = "org.zdoom.GZDoom"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("WAD file"),
            "help": _("The game data, commonly called a WAD file."),
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Arguments"),
            "help": _("Command line arguments used when launching the game."),
        },
        {
            "option": "files",
            "type": "multiple_file",
            "label": _("PWAD files"),
            "help": _("Used to load one or more PWAD files which generally contain user-created levels."),
        },
        {
            "option": "warp",
            "type": "string",
            "label": _("Warp to map"),
            "help": _("Starts the game on the given map."),
        },
        {
            "option": "savedir",
            "type": "directory",
            "label": _("Save path"),
            "warn_if_non_writable_parent": True,
            "help": _("User-specified path where save files should be located."),
        },
    ]
    runner_options = [
        {"option": "2", "label": _("Pixel Doubling"), "type": "bool", "default": False},
        {"option": "4", "label": _("Pixel Quadrupling"), "type": "bool", "default": False},
        {
            "option": "nostartup",
            "label": _("Disable Startup Screens"),
            "type": "bool",
            "default": False,
        },
        {
            "option": "skill",
            "label": _("Skill"),
            "type": "choice",
            "default": "",
            "choices": {
                (_("None"), ""),
                (_("I'm Too Young To Die (1)"), "1"),
                (_("Hey, Not Too Rough (2)"), "2"),
                (_("Hurt Me Plenty (3)"), "3"),
                (_("Ultra-Violence (4)"), "4"),
                (_("Nightmare! (5)"), "5"),
            },
        },
        {
            "option": "config",
            "label": _("Config file"),
            "type": "file",
            "help": _(
                "Used to load a user-created configuration file. If specified, "
                "the file must contain the wad directory list or launch will fail."
            ),
        },
    ]

    def prelaunch(self):
        if not LINUX_SYSTEM.get_soundfonts():
            logger.warning("FluidSynth is not installed, you might not have any music")

    @property
    def working_dir(self):
        wad = self.game_config.get("main_file")
        if wad:
            return os.path.dirname(os.path.expanduser(wad))
        wad_files = self.game_config.get("files")
        if wad_files:
            return os.path.dirname(os.path.expanduser(wad_files[0]))

    def play(self):  # noqa: C901
        command = self.get_command()

        resolution = self.runner_config.get("resolution")
        if resolution:
            if resolution == "desktop":
                width, height = display.DISPLAY_MANAGER.get_current_resolution()
            else:
                width, height = resolution.split("x")
            command.append("-width")
            command.append(width)
            command.append("-height")
            command.append(height)

        # Append any boolean options.
        bool_options = ["2", "4", "nostartup"]
        for option in bool_options:
            if self.runner_config.get(option):
                command.append("-%s" % option)

        # Append the skill level.
        skill = self.runner_config.get("skill")
        if skill:
            command.append("-skill")
            command.append(skill)

        # Append directory for configuration file, if provided.
        config = self.runner_config.get("config")
        if config:
            command.append("-config")
            command.append(config)

        # Append the warp arguments.
        warp = self.game_config.get("warp")
        if warp:
            command.append("-warp")
            for warparg in warp.split(" "):
                command.append(warparg)

        # Append directory for save games, if provided.
        savedir = self.game_config.get("savedir")
        if savedir:
            command.append("-savedir")
            command.append(savedir)

        # Append the wad file to load, if provided.
        wad = self.game_config.get("main_file")
        if wad:
            command.append("-iwad")
            command.append(wad)

        # Append the pwad files to load, if provided.
        files = self.game_config.get("files") or []
        pwads = [f for f in files if f.lower().endswith(".wad") or f.lower().endswith(".pk3")]
        deh = [f for f in files if f.lower().endswith(".deh")]
        bex = [f for f in files if f.lower().endswith(".bex")]
        if deh:
            command.append("-deh")
            command.append(deh[0])
        if bex:
            command.append("-bex")
            command.append(bex[0])
        if pwads:
            command.append("-file")
            for pwad in pwads:
                command.append(pwad)

        # Append additional arguments, if provided.
        args = self.game_config.get("args") or ""
        for arg in split_arguments(args):
            command.append(arg)

        return {"command": command}



"""libretro runner"""

import os
from gettext import gettext as _
from operator import itemgetter
from zipfile import ZipFile

import requests

from lutris import settings
from lutris.config import LutrisConfig
from lutris.exceptions import GameConfigError, MissingBiosError, MissingGameExecutableError, UnspecifiedVersionError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.libretro import RetroConfig
from lutris.util.log import logger
from lutris.util.retroarch.firmware import get_firmware, scan_firmware_directory

RETROARCH_DIR = os.path.join(settings.RUNNER_DIR, "retroarch")


def get_default_config_path(path):
    return os.path.join(RETROARCH_DIR, path)


def get_libretro_cores():
    cores = []
    if not os.path.exists(RETROARCH_DIR):
        return []

    # Get core identifiers from info dir
    info_path = os.path.join(RETROARCH_DIR, "info")
    if not os.path.exists(info_path):
        req = requests.get("http://buildbot.libretro.com/assets/frontend/info.zip", allow_redirects=True, timeout=5)
        if req.status_code == requests.codes.ok:  # pylint: disable=no-member
            with open(os.path.join(RETROARCH_DIR, "info.zip"), "wb") as info_zip:
                info_zip.write(req.content)
            with ZipFile(os.path.join(RETROARCH_DIR, "info.zip"), "r") as info_zip:
                info_zip.extractall(info_path)
        else:
            logger.error("Error retrieving libretro info archive from server: %s - %s", req.status_code, req.reason)
            return []
    # Parse info files to fetch display name and platform/system
    for info_file in os.listdir(info_path):
        if "_libretro.info" not in info_file:
            continue
        core_identifier = info_file.replace("_libretro.info", "")
        core_config = RetroConfig(os.path.join(info_path, info_file))
        if "categories" in core_config.keys() and "Emulator" in core_config["categories"]:
            core_label = core_config["display_name"] or ""
            core_system = core_config["systemname"] or ""
            cores.append((core_label, core_identifier, core_system))
    cores.sort(key=itemgetter(0))
    return cores


# List of supported libretro cores
# First element is the human readable name for the core with the platform's short name
# Second element is the core identifier
# Third element is the platform's long name
LIBRETRO_CORES = get_libretro_cores()


def get_core_choices():
    return [(core[0], core[1]) for core in LIBRETRO_CORES]


class libretro(Runner):
    human_name = _("Libretro")
    description = _("Multi-system emulator")
    runnable_alone = True
    runner_executable = "retroarch/retroarch"
    flatpak_id = "org.libretro.RetroArch"

    game_options = [
        {"option": "main_file", "type": "file", "label": _("ROM file")},
        {
            "option": "core",
            "type": "choice",
            "label": _("Core"),
            "choices": get_core_choices(),
        },
    ]

    runner_options = [
        {
            "option": "config_file",
            "type": "file",
            "label": _("Config file"),
            "default": os.path.join(RETROARCH_DIR, "retroarch.cfg"),
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": True,
        },
        {
            "option": "verbose",
            "type": "bool",
            "label": _("Verbose logging"),
            "default": False,
        },
    ]

    @property
    def directory(self):
        return os.path.join(settings.RUNNER_DIR, "retroarch")

    @property
    def platforms(self):
        return [core[2] for core in LIBRETRO_CORES]

    def get_platform(self):
        game_core = self.game_config.get("core")
        if not game_core:
            logger.warning("Game don't have a core set")
            return
        for core in LIBRETRO_CORES:
            if core[1] == game_core:
                return core[2]
        logger.warning("'%s' not found in Libretro cores", game_core)
        return ""

    def get_core_path(self, core):
        """Return the path of a core from libretro's runner only"""
        lutris_cores_folder = get_default_config_path("cores")
        core_filename = "{}_libretro.so".format(core)
        lutris_core = os.path.join(lutris_cores_folder, core_filename)
        return lutris_core

    def get_version(self, use_default=True):
        return self.game_config["core"]

    def is_installed(self, flatpak_allowed: bool = True, core=None) -> bool:
        if not core and self.has_explicit_config and self.game_config.get("core"):
            core = self.game_config["core"]
        if not core or self.runner_config.get("runner_executable"):
            return super().is_installed(flatpak_allowed=flatpak_allowed)
        is_core_installed = system.path_exists(self.get_core_path(core))
        return super().is_installed(flatpak_allowed=flatpak_allowed) and is_core_installed

    def is_installed_for(self, interpreter):
        core = interpreter.installer.script["game"].get("core")
        return self.is_installed(core=core)

    def get_installer_runner_version(self, installer, use_runner_config: bool = True) -> str:
        version = installer.script["game"].get("core")
        if not version:
            raise UnspecifiedVersionError(_("The installer does not specify the libretro 'core' version."))
        return version

    def install(self, install_ui_delegate, version=None, callback=None):
        captured_super = super()  # super() does not work inside install_core()

        def install_core():
            if not version:
                if callback:
                    callback()
            else:
                captured_super.install(install_ui_delegate, version, callback)

        if not super().is_installed():
            captured_super.install(install_ui_delegate, version=None, callback=install_core)
        else:
            captured_super.install(install_ui_delegate, version, callback)

    def get_run_data(self):
        return {
            "command": self.get_command() + self.get_runner_parameters(),
            "env": self.get_env(),
        }

    def get_config_file(self):
        return self.runner_config.get("config_file") or os.path.join(RETROARCH_DIR, "retroarch.cfg")

    @staticmethod
    def get_system_directory(retro_config):
        """Return the system directory used for storing BIOS and firmwares."""
        system_directory = retro_config["system_directory"]
        if not system_directory or system_directory == "default":
            system_directory = get_default_config_path("system")
        return os.path.expanduser(system_directory)

    def prelaunch(self):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        config_file = self.get_config_file()
        # TODO: review later
        # Create retroarch.cfg if it doesn't exist.
        if not system.path_exists(config_file):
            with open(config_file, "w", encoding="utf-8") as f:
                f.write("# Lutris RetroArch Configuration")
                f.close()

            # Build the default config settings.
            retro_config = RetroConfig(config_file)
            retro_config["libretro_directory"] = get_default_config_path("cores")
            retro_config["libretro_info_path"] = get_default_config_path("info")
            retro_config["content_database_path"] = get_default_config_path("database/rdb")
            retro_config["cheat_database_path"] = get_default_config_path("database/cht")
            retro_config["cursor_directory"] = get_default_config_path("database/cursors")
            retro_config["screenshot_directory"] = get_default_config_path("screenshots")
            retro_config["input_remapping_directory"] = get_default_config_path("remaps")
            retro_config["video_shader_dir"] = get_default_config_path("shaders")
            retro_config["core_assets_directory"] = get_default_config_path("downloads")
            retro_config["thumbnails_directory"] = get_default_config_path("thumbnails")
            retro_config["playlist_directory"] = get_default_config_path("playlists")
            retro_config["joypad_autoconfig_dir"] = get_default_config_path("autoconfig")
            retro_config["rgui_config_directory"] = get_default_config_path("config")
            retro_config["overlay_directory"] = get_default_config_path("overlay")
            retro_config["assets_directory"] = get_default_config_path("assets")
            retro_config["system_directory"] = get_default_config_path("system")
            retro_config.save()
        else:
            retro_config = RetroConfig(config_file)

        core = self.game_config.get("core")
        info_file = os.path.join(RETROARCH_DIR, f"info/{core}_libretro.info")
        if system.path_exists(info_file):
            retro_config = RetroConfig(info_file)
            try:
                firmware_count = int(retro_config["firmware_count"])
            except (ValueError, TypeError):
                firmware_count = 0
            system_path = self.get_system_directory(retro_config)
            notes = str(retro_config["notes"] or "")
            checksums = {}
            if notes.startswith("(!)"):
                parts = notes.split("|")
                for part in parts:
                    try:
                        filename, checksum = part.split(" (md5): ")
                    except ValueError:
                        logger.warning("Unable to parse firmware info: %s", notes)
                        continue
                    filename = filename.replace("(!) ", "")
                    checksums[filename] = checksum

            # If this requires firmware, confirm we have the firmware folder configured in the first place
            # then rescan it in case the user added anything since the last time they changed it
            if firmware_count > 0:
                lutris_config = LutrisConfig()
                firmware_directory = lutris_config.raw_system_config.get("bios_path")
                if not firmware_directory:
                    raise MissingBiosError(
                        _("The emulator files BIOS location must be configured in the Preferences dialog.")
                    )
                scan_firmware_directory(firmware_directory)

            for index in range(firmware_count):
                required_firmware_filename = retro_config["firmware%d_path" % index]
                required_firmware_path = os.path.join(system_path, required_firmware_filename)
                required_firmware_name = required_firmware_filename.split("/")[-1]
                required_firmware_checksum = checksums.get(required_firmware_name)
                if system.path_exists(required_firmware_path):
                    if required_firmware_checksum:
                        checksum = system.get_md5_hash(required_firmware_path)
                        if checksum == required_firmware_checksum:
                            checksum_status = "Checksum good"
                        else:
                            checksum_status = "Checksum failed"
                    else:
                        checksum_status = "No checksum info"
                    logger.info("Firmware '%s' found (%s)", required_firmware_filename, checksum_status)
                else:
                    logger.warning("Firmware '%s' not found!", required_firmware_filename)
                    if required_firmware_checksum:
                        get_firmware(required_firmware_name, required_firmware_checksum, system_path)

    def get_runner_parameters(self):
        parameters = []

        # Fullscreen
        fullscreen = self.runner_config.get("fullscreen")
        if fullscreen:
            parameters.append("--fullscreen")

        # Verbose
        verbose = self.runner_config.get("verbose")
        if verbose:
            parameters.append("--verbose")

        parameters.append("--config={}".format(self.get_config_file()))
        return parameters

    def play(self):
        command = self.get_command() + self.get_runner_parameters()

        # Core
        core = self.game_config.get("core")
        if not core:
            raise GameConfigError(_("No core has been selected for this game"))
        command.append("--libretro={}".format(self.get_core_path(core)))

        # Main file
        file = self.game_config.get("main_file")
        if not file:
            raise GameConfigError(_("No game file specified"))
        if not system.path_exists(file):
            raise MissingGameExecutableError(filename=file)
        command.append(file)
        return {"command": command}



"""Options list for system config."""

import os
from collections import OrderedDict
from gettext import gettext as _
from gettext import pgettext as C_

from lutris import runners
from lutris.util import linux, system
from lutris.util.display import DISPLAY_MANAGER, SCREEN_SAVER_INHIBITOR, is_compositing_enabled, is_display_x11
from lutris.util.graphics.gpu import GPUS


def get_resolution_choices():
    """Return list of available resolutions as label, value tuples
    suitable for inclusion in drop-downs.
    """
    resolutions = DISPLAY_MANAGER.get_resolutions()
    resolution_choices = list(zip(resolutions, resolutions))
    resolution_choices.insert(0, (_("Keep current"), "off"))
    return resolution_choices


def get_locale_choices():
    """Return list of available locales as label, value tuples
    suitable for inclusion in drop-downs.
    """
    return [
        (C_("locale", "System"), ""),
        (_("Chinese"), "zh_CN.utf8"),
        (_("Croatian"), "hr_HR.utf8"),
        (_("Dutch"), "nl_NL.utf8"),
        (_("English"), "en_US.utf8"),
        (_("Finnish"), "fi_FI.utf8"),
        (_("French"), "fr_FR.utf8"),
        (_("Georgian"), "ka_GE.utf8"),
        (_("German"), "de_DE.utf8"),
        (_("Greek"), "el_GR.utf8"),
        (_("Italian"), "it_IT.utf8"),
        (_("Japanese"), "ja_JP.utf8"),
        (_("Korean"), "ko_KR.utf8"),
        (_("Portuguese (Brazilian)"), "pt_BR.utf8"),
        (_("Polish"), "pl_PL.utf8"),
        (_("Russian"), "ru_RU.utf8"),
        (_("Spanish"), "es_ES.utf8"),
        (_("Turkish"), "tr_TR.utf8"),
    ]


def get_gpu_list():
    choices = [(_("Auto"), "")]
    for card, gpu in GPUS.items():
        choices.append((gpu.short_name, card))
    return choices


def get_output_choices():
    """Return list of outputs for drop-downs"""
    displays = DISPLAY_MANAGER.get_display_names()
    output_choices = list(zip(displays, displays))
    output_choices.insert(0, (_("Off"), "off"))
    output_choices.insert(1, (_("Primary"), "primary"))
    return output_choices


def get_output_list():
    """Return a list of output with their index.
    This is used to indicate to SDL 1.2 which monitor to use.
    """
    choices = [(_("Off"), "off")]
    displays = DISPLAY_MANAGER.get_display_names()
    for index, output in enumerate(displays):
        # Display name can't be used because they might not be in the right order
        # Using DISPLAYS to get the number of connected monitors
        choices.append((output, str(index)))
    return choices


system_options = [  # pylint: disable=invalid-name
    {
        "section": _("Lutris"),
        "option": "game_path",
        "type": "directory",
        "label": _("Default installation folder"),
        "warn_if_non_writable_parent": True,
        "default": os.path.expanduser("~/Games"),
        "scope": ["runner", "system"],
        "help": _("The default folder where you install your games."),
    },
    {
        "section": _("Lutris"),
        "option": "disable_runtime",
        "type": "bool",
        "label": _("Disable Lutris Runtime"),
        "default": False,
        "help": _(
            "The Lutris Runtime loads some libraries before running the "
            "game, which can cause some incompatibilities in some cases. "
            "Check this option to disable it."
        ),
    },
    {
        "section": _("Lutris"),
        "option": "prefer_system_libs",
        "type": "bool",
        "label": _("Prefer system libraries"),
        "default": True,
        "help": _("When the runtime is enabled, prioritize the system libraries over the provided ones."),
    },
    {
        "section": _("Display"),
        "option": "gpu",
        "type": "choice",
        "label": _("GPU"),
        "choices": get_gpu_list,
        "default": "",
        "condition": lambda: len(GPUS) > 1,
        "help": _("GPU to use to run games"),
    },
    {
        "section": _("Display"),
        "option": "mangohud",
        "type": "bool",
        "label": _("FPS counter (MangoHud)"),
        "default": False,
        "condition": system.can_find_executable("mangohud"),
        "help": _("Display the game's FPS + other information. Requires MangoHud to be installed."),
    },
    {
        "section": _("Display"),
        "option": "reset_desktop",
        "type": "bool",
        "label": _("Restore resolution on game exit"),
        "default": False,
        "available": is_display_x11,
        "advanced": True,
        "help": _(
            "Some games don't restore your screen resolution when \n"
            "closed or when they crash. This is when this option comes \n"
            "into play to save your bacon."
        ),
    },
    {
        "section": _("Display"),
        "option": "disable_compositor",
        "label": _("Disable desktop effects"),
        "type": "bool",
        "default": False,
        "advanced": True,
        "available": is_display_x11,
        "condition": is_compositing_enabled(),
        "help": _("Disable desktop effects while game is running, reducing stuttering and increasing performance"),
    },
    {
        "section": _("Display"),
        "option": "disable_screen_saver",
        "label": _("Prevent sleep"),
        "type": "bool",
        "default": SCREEN_SAVER_INHIBITOR is not None,
        "advanced": True,
        "condition": SCREEN_SAVER_INHIBITOR is not None,
        "help": _("Prevents the computer from suspending when a game is running."),
    },
    {
        "section": _("Display"),
        "option": "sdl_video_fullscreen",
        "type": "choice",
        "label": _("SDL 1.2 Fullscreen Monitor"),
        "choices": get_output_list,
        "default": "off",
        "available": is_display_x11,
        "advanced": True,
        "help": _(
            "Hint SDL 1.2 games to use a specific monitor when going "
            "fullscreen by setting the SDL_VIDEO_FULLSCREEN "
            "environment variable"
        ),
    },
    {
        "section": _("Display"),
        "option": "display",
        "type": "choice",
        "label": _("Turn off monitors except"),
        "choices": get_output_choices,
        "default": "off",
        "available": is_display_x11,
        "advanced": True,
        "help": _(
            "Only keep the selected screen active while the game is "
            "running. \n"
            "This is useful if you have a dual-screen setup, and are \n"
            "having display issues when running a game in fullscreen."
        ),
    },
    {
        "section": _("Display"),
        "option": "resolution",
        "type": "choice",
        "label": _("Switch resolution to"),
        "advanced": True,
        "available": is_display_x11,
        "choices": get_resolution_choices,
        "default": "off",
        "help": _("Switch to this screen resolution while the game is running."),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope",
        "type": "bool",
        "label": _("Enable Gamescope"),
        "default": False,
        "condition": system.can_find_executable("gamescope") and linux.LINUX_SYSTEM.nvidia_gamescope_support(),
        "help": _("Use gamescope to draw the game window isolated from your desktop.\nToggle fullscreen: Super + F"),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_hdr",
        "type": "bool",
        "label": _("Enable HDR (Experimental)"),
        "advanced": False,
        "default": False,
        "conditional_on": "gamescope",
        "condition": bool(system.can_find_executable("gamescope")),
        "help": _("Enable HDR for games that support it.\nRequires Plasma 6 and VK_hdr_layer."),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_force_grab_cursor",
        "type": "bool",
        "label": _("Relative Mouse Mode"),
        "advanced": True,
        "default": False,
        "conditional_on": "gamescope",
        "condition": bool(system.can_find_executable("gamescope")),
        "help": _(
            "Always use relative mouse mode instead of flipping\n"
            "dependent on cursor visibility\n"
            "Can help with games where the player's camera faces the floor"
        ),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_output_res",
        "type": "choice_with_entry",
        "label": _("Output Resolution"),
        "choices": DISPLAY_MANAGER.get_resolutions,
        "advanced": True,
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _(
            "Set the resolution used by gamescope.\n"
            "Resizing the gamescope window will update these settings.\n"
            "\n"
            "<b>Custom Resolutions:</b> (width)x(height)"
        ),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_game_res",
        "type": "choice_with_entry",
        "label": _("Game Resolution"),
        "choices": DISPLAY_MANAGER.get_resolutions,
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _("Set the maximum resolution used by the game.\n\n<b>Custom Resolutions:</b> (width)x(height)"),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_window_mode",
        "label": _("Window Mode"),
        "type": "choice",
        "choices": (
            (_("Fullscreen"), "-f"),
            (_("Windowed"), ""),
            (_("Borderless"), "-b"),
        ),
        "default": "-f",
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _("Run gamescope in fullscreen, windowed or borderless mode\nToggle fullscreen : Super + F"),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_fsr_sharpness",
        "label": _("FSR Level"),
        "advanced": True,
        "type": "string",
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _(
            "Use AMD FidelityFX™ Super Resolution 1.0 for upscaling.\nUpscaler sharpness from 0 (max) to 20 (min)."
        ),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_fps_limiter",
        "label": _("Framerate Limiter"),
        "advanced": False,
        "type": "string",
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _("Set a frame-rate limit for gamescope specified in frames per second."),
    },
    {
        "section": _("Gamescope"),
        "option": "gamescope_flags",
        "label": _("Custom Settings"),
        "advanced": True,
        "type": "string",
        "conditional_on": "gamescope",
        "condition": system.can_find_executable("gamescope"),
        "help": _(
            "Set additional flags for gamescope (if available).\nSee 'gamescope --help' for a full list of options."
        ),
    },
    {
        "section": _("CPU"),
        "option": "single_cpu",
        "type": "bool",
        "label": _("Restrict number of cores used"),
        "default": False,
        "help": _("Restrict the game to a maximum number of CPU cores."),
    },
    {
        "section": _("CPU"),
        "option": "limit_cpu_count",
        "type": "string",
        "label": _("Restrict number of cores to"),
        "default": "1",
        "conditional_on": "single_cpu",
        "help": _("Maximum number of CPU cores to be used, if 'Restrict number of cores used' is turned on."),
    },
    {
        "section": _("CPU"),
        "option": "gamemode",
        "type": "bool",
        "default": linux.LINUX_SYSTEM.gamemode_available(),
        "condition": linux.LINUX_SYSTEM.gamemode_available(),
        "label": _("Enable Feral GameMode"),
        "help": _("Request a set of optimisations be temporarily applied to the host OS"),
    },
    {
        "section": _("Audio"),
        "option": "pulse_latency",
        "type": "bool",
        "label": _("Reduce PulseAudio latency"),
        "default": False,
        "advanced": True,
        "condition": system.can_find_executable("pulseaudio") or system.can_find_executable("pipewire-pulse"),
        "help": _("Set the environment variable PULSE_LATENCY_MSEC=60 to improve audio quality on some games"),
    },
    {
        "section": _("Input"),
        "option": "use_us_layout",
        "type": "bool",
        "label": _("Switch to US keyboard layout"),
        "default": False,
        "available": is_display_x11,
        "advanced": True,
        "help": _("Switch to US keyboard QWERTY layout while game is running"),
    },
    {
        "section": _("Input"),
        "option": "antimicro_config",
        "type": "file",
        "label": _("AntiMicroX Profile"),
        "advanced": True,
        "help": _("Path to an AntiMicroX profile file"),
    },
    {
        "section": _("Input"),
        "option": "sdl_gamecontrollerconfig",
        "type": "string",
        "label": _("SDL2 gamepad mapping"),
        "advanced": True,
        "help": _(
            "SDL_GAMECONTROLLERCONFIG mapping string or path to a custom gamecontrollerdb.txt file containing mappings."
        ),
    },
    {
        "section": _("Text based games"),
        "option": "terminal",
        "label": _("CLI mode"),
        "type": "bool",
        "default": False,
        "advanced": True,
        "help": _(
            "Enable a terminal for text-based games. "
            "Only useful for ASCII based games. May cause issues with graphical games."
        ),
    },
    {
        "section": _("Text based games"),
        "option": "terminal_app",
        "label": _("Text based games emulator"),
        "type": "choice_with_entry",
        "choices": linux.get_terminal_apps,
        "default": linux.get_default_terminal(),
        "advanced": True,
        "help": _(
            "The terminal emulator used with the CLI mode. "
            "Choose from the list of detected terminal apps or enter "
            "the terminal's command or path."
        ),
    },
    {
        "section": _("Game execution"),
        "option": "env",
        "type": "mapping",
        "label": _("Environment variables"),
        "help": _("Environment variables loaded at run time"),
    },
    {
        "section": _("Game execution"),
        "option": "locale",
        "type": "choice_with_entry",
        "label": _("Locale"),
        "choices": (get_locale_choices),
        "default": "",
        "advanced": False,
        "help": _("Can be used to force certain locale for an app. Fixes encoding issues in legacy software."),
    },
    {
        "section": _("Game execution"),
        "option": "prefix_command",
        "type": "string",
        "label": _("Command prefix"),
        "advanced": True,
        "help": _("Command line instructions to add in front of the game's execution command."),
    },
    {
        "section": _("Game execution"),
        "option": "manual_command",
        "type": "file",
        "label": _("Manual script"),
        "advanced": True,
        "help": _("Script to execute from the game's contextual menu"),
    },
    {
        "section": _("Game execution"),
        "option": "prelaunch_command",
        "type": "command_line",
        "label": _("Pre-launch script"),
        "advanced": True,
        "help": _("Script to execute before the game starts"),
    },
    {
        "section": _("Game execution"),
        "option": "prelaunch_wait",
        "type": "bool",
        "label": _("Wait for pre-launch script completion"),
        "advanced": True,
        "default": False,
        "conditional_on": "prelaunch_command",
        "help": _("Run the game only once the pre-launch script has exited"),
    },
    {
        "section": _("Game execution"),
        "option": "postexit_command",
        "type": "command_line",
        "label": _("Post-exit script"),
        "advanced": True,
        "help": _("Script to execute when the game exits"),
    },
    {
        "section": _("Game execution"),
        "option": "include_processes",
        "type": "string",
        "label": _("Include processes"),
        "advanced": True,
        "help": _(
            "What processes to include in process monitoring. "
            "This is to override the built-in exclude list.\n"
            "Space-separated list, processes including spaces "
            "can be wrapped in quotation marks."
        ),
    },
    {
        "section": _("Game execution"),
        "option": "exclude_processes",
        "type": "string",
        "label": _("Exclude processes"),
        "advanced": True,
        "help": _(
            "What processes to exclude in process monitoring. "
            "For example background processes that stick around "
            "after the game has been closed.\n"
            "Space-separated list, processes including spaces "
            "can be wrapped in quotation marks."
        ),
    },
    {
        "section": _("Game execution"),
        "option": "killswitch",
        "type": "string",
        "label": _("Killswitch file"),
        "advanced": True,
        "help": _(
            "Path to a file which will stop the game when deleted \n"
            "(usually /dev/input/js0 to stop the game on joystick "
            "unplugging)"
        ),
    },
]


def with_runner_overrides(runner_slug):
    """Return system options updated with overrides from given runner."""
    options = system_options
    try:
        runner = runners.import_runner(runner_slug)()
    except runners.InvalidRunnerError:
        return options
    if runner.system_options_override:
        opts_dict = OrderedDict((opt["option"], opt) for opt in options)
        for option in runner.system_options_override:
            key = option["option"]
            if opts_dict.get(key):
                opts_dict[key] = opts_dict[key].copy()
                opts_dict[key].update(option)
            else:
                opts_dict[key] = option
        options = list(opts_dict.values())
    return options



"""Generic service utilities"""

import os
import shutil
from gettext import gettext as _
from pathlib import Path
from typing import Any, Dict, List

from gi.repository import Gio

from lutris import api, settings
from lutris.api import get_game_installers
from lutris.config import write_game_config
from lutris.database import sql
from lutris.database.games import add_game, get_game_by_field, get_game_for_service, get_games
from lutris.database.services import ServiceGameCollection
from lutris.game import GAME_UPDATED, Game
from lutris.gui.dialogs import NoticeDialog
from lutris.gui.dialogs.webconnect_dialog import WebConnectDialog
from lutris.gui.views.media_loader import download_media
from lutris.gui.widgets import NotificationSource
from lutris.gui.widgets.utils import BANNER_SIZE, ICON_SIZE
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.busy import BusyAsyncCall
from lutris.util.cookies import WebkitCookieJar
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger
from lutris.util.strings import slugify


class AuthTokenExpiredError(Exception):
    """Exception raised when a token is no longer valid; the sidebar will
    log-out and log-in again in response to this rather than reporting it."""


class LutrisBanner(ServiceMedia):
    service = "lutris"
    size = BANNER_SIZE
    dest_path = settings.BANNER_PATH
    file_patterns = ["%s.jpg", "%s.png"]
    api_field = "banner"


class LutrisIcon(LutrisBanner):
    size = ICON_SIZE
    dest_path = settings.ICON_PATH
    file_patterns = ["lutris_%s.png"]
    api_field = "icon"
    can_be_fallback = False

    @property
    def custom_media_storage_size(self):
        return (128, 128)

    def run_system_update_desktop_icons(self):
        system.update_desktop_icons()

    def get_fallback_media_path(self, services):
        return None


class LutrisCoverart(ServiceMedia):
    service = "lutris"
    size = (264, 352)
    file_patterns = ["%s.jpg", "%s.png"]
    dest_path = settings.COVERART_PATH
    api_field = "coverart"

    @property
    def config_ui_size(self):
        return (66, 88)


class LutrisCoverartMedium(LutrisCoverart):
    size = (176, 234)


SERVICE_GAMES_LOADING = NotificationSource()
SERVICE_GAMES_LOADED = NotificationSource()
SERVICE_LOGIN = NotificationSource()
SERVICE_LOGOUT = NotificationSource()


class BaseService:
    """Base class for local services"""

    id = NotImplemented
    _matcher = None
    has_extras = False
    name = NotImplemented
    description = ""
    icon = NotImplemented
    online = False
    local = False
    drm_free = False  # DRM free games can be added to Lutris from an existing install
    client_installer = None  # ID of a script needed to install the client used by the service
    scripts = {}  # Mapping of Javascript snippets to handle redirections during auth
    medias = {}
    extra_medias = {}
    default_format = "icon"
    is_loading = False

    @property
    def matcher(self):
        if self._matcher:
            return self._matcher
        return self.id

    def run(self, launch_ui_delegate):
        """Launch the game client"""
        launcher = self.get_launcher()
        if launcher:
            launcher.launch(launch_ui_delegate)

    def is_launchable(self):
        if self.client_installer:
            return bool(get_game_by_field(self.client_installer, "slug"))
        return False

    def get_launcher(self):
        if not self.client_installer:
            return None
        db_launcher = get_game_by_field(self.client_installer, "slug")
        if db_launcher:
            return Game(db_launcher["id"])

    def is_launcher_installed(self):
        launcher = self.get_launcher()
        if not launcher:
            return False
        return launcher.is_installed

    def start_reload(self, reloaded_callback):
        """Refresh the service's games, asynchronously. This raises signals, but
        does so on the main thread- and runs the reload on a worker thread. It calls
        reloaded_callback when done, passing any error (or None on success)"""

        def do_reload():
            if self.is_loading:
                logger.warning("'%s' games are already loading", self.name)
                return

            try:
                self.is_loading = True

                self.wipe_game_cache()
                self.load()
                self.load_icons()
                self.add_installed_games()
                logger.debug("'%s' games reloaded", self.name)
            finally:
                self.is_loading = False

        def reload_cb(_result, error):
            logger.debug("Reload callback")
            SERVICE_GAMES_LOADED.fire(self)
            reloaded_callback(error)

        SERVICE_GAMES_LOADING.fire(self)
        AsyncCall(do_reload, reload_cb)

    def load(self):
        logger.warning("Load method not implemented")

    def load_icons(self):
        """Download all game media from the service"""
        logger.debug("Loading icons...")
        all_medias = self.medias.copy()
        all_medias.update(self.extra_medias)

        service_medias = [media_type() for media_type in all_medias.values()]

        # Download icons
        for service_media in service_medias:
            media_urls = service_media.get_media_urls()
            download_media(media_urls, service_media)

        # Process icons
        for service_media in service_medias:
            service_media.render()

    def wipe_game_cache(self):
        logger.debug("Deleting games from service-games for %s", self.id)
        sql.db_delete(settings.DB_PATH, "service_games", "service", self.id)

    def get_update_installers(self, db_game):
        return []

    def generate_installer(self, db_game: Dict[str, Any]) -> Dict[str, Any]:
        """Used to generate an installer from the data returned from the services"""
        return {}

    def generate_installers(self, db_game: Dict[str, Any]) -> List[dict]:
        """Used to generate a list of installers to choose from, from the data returned from the services
        By default this is just the installer from generate_installer(), and if overridden to return
        more than one, then generate_installer must be overridden ti pick a default installer."""
        installer = self.generate_installer(db_game)
        return [installer] if installer else []

    def install_from_api(self, db_game, appid=None):
        """Install a game, using the API or generate_installer() to obtain the installer."""
        if not appid:
            appid = db_game["appid"]

        def on_installers_ready(service_installers, error):
            if error:
                raise error  # bounce any error off the backstop

            if not service_installers:
                service_installers = self.generate_installers(db_game)
            application = Gio.Application.get_default()
            application.show_installer_window(service_installers, service=self, appid=appid)

        BusyAsyncCall(self.get_installers_from_api, on_installers_ready, appid)

    def get_installer_files(self, installer, installer_file_id, selected_extras):
        """Used to obtains the content files from the service, when an 'N/A' file is left in
        the installer. This handles 'extras', and must return a tuple; first a list of
        InstallerFile or InstallerFileCollection objects that are for the files themselves,
        and then a list of such objects for the extras. This separation allows us to generate
        extra installer script steps to move the extras in."""
        return [], []

    def match_game(self, service_game, lutris_game):
        """Match a service game to a lutris game referenced by its slug"""
        if not service_game:
            return
        sql.db_update(
            settings.DB_PATH,
            "service_games",
            {"lutris_slug": lutris_game["slug"]},
            conditions={"appid": service_game["appid"], "service": self.id},
        )
        unmatched_lutris_games = get_games(
            searches={"installer_slug": self.matcher},
            filters={"slug": lutris_game["slug"]},
            excludes={"service": self.id},
        )
        for game in unmatched_lutris_games:
            logger.debug("Updating unmatched game %s", game)
            sql.db_update(
                settings.DB_PATH,
                "games",
                {"service": self.id, "service_id": service_game["appid"]},
                conditions={"id": game["id"]},
            )

    def match_games(self):
        """Matching of service games to lutris games"""
        service_games = {str(game["appid"]): game for game in ServiceGameCollection.get_for_service(self.id)}
        lutris_games = api.get_api_games(list(service_games.keys()), service=self.id)
        for lutris_game in lutris_games:
            for provider_game in lutris_game["provider_games"]:
                if provider_game["service"] != self.id:
                    continue
                self.match_game(service_games.get(provider_game["slug"]), lutris_game)
        unmatched_service_games = get_games(searches={"installer_slug": self.matcher}, excludes={"service": self.id})
        for lutris_game in api.get_api_games(game_slugs=[g["slug"] for g in unmatched_service_games]):
            for provider_game in lutris_game["provider_games"]:
                if provider_game["service"] != self.id:
                    continue
                self.match_game(service_games.get(provider_game["slug"]), lutris_game)

    def match_existing_game(self, db_games, appid, no_signal=False):
        """Checks if a game is already installed and populates the service info"""
        for _game in db_games:
            logger.debug("Matching %s with existing install: %s", appid, _game)
            game = Game(_game["id"])
            game.appid = appid
            game.service = self.id
            game.save(no_signal=no_signal)
            service_game = ServiceGameCollection.get_game(self.id, appid)
            sql.db_update(settings.DB_PATH, "service_games", {"lutris_slug": game.slug}, {"id": service_game["id"]})
            return game

    def get_installers_from_api(self, appid):
        """Query the lutris API for an appid and get existing installers for the service"""
        lutris_games = api.get_api_games([appid], service=self.id)
        service_installers = []
        if lutris_games:
            lutris_game = lutris_games[0]
            installers = get_game_installers(lutris_game["slug"])
            for installer in installers:
                if self.matcher in installer["version"].lower():
                    service_installers.append(installer)
        return service_installers

    def get_installed_slug(self, db_game):
        """Returns the slug the game will have after installation, by default. This
        is Lutris's slug, not the one for the service. By default, we derive it from
        the Game's name."""
        return slugify(db_game["name"])

    def get_installed_runner_name(self, db_game):
        """Returns the name of the runner this game will have after installation, or
        blank if this is not known."""
        return ""

    def get_service_installers(self, db_game, update):
        appid = db_game["appid"]
        if update:
            service_installers = self.get_update_installers(db_game)
        else:
            service_installers = self.get_installers_from_api(appid)
        # Check if the game is not already installed
        for service_installer in service_installers:
            existing_game = self.match_existing_game(
                get_games(filters={"installer_slug": service_installer["slug"], "installed": "1"}),
                appid,
                no_signal=True,  # we're on a thread here, signals can crash us!
            )
            if existing_game:
                logger.debug("Found existing game, aborting install")
                return None, None, existing_game
        installers = self.generate_installers(db_game) if not update else []
        if installers:
            if service_installers:
                for installer in installers:
                    installer["version"] = installer["version"] + " (auto-generated)"
            service_installers.extend(installers)
        if not service_installers:
            logger.error("No installer found for %s", db_game)
        return service_installers, db_game, None

    def install(self, db_game, update=False):
        """Install a service game, or starts the installer of the game.

        Args:
            db_game (dict or str): Database fields of the game to add, or (for Lutris service only
                the slug of the game.)

        Returns:
            str: The slug of the game that was installed, to be run. None if the game should not be
                run now. Many installers start from here, but continue running after this returns;
                they return None.
        """
        logger.debug("Installing %s from service %s", db_game["appid"], self.id)
        # Local services (aka game libraries that don't require any type of online interaction) can
        # be added without going through an install dialog.
        if self.local:
            return self.simple_install(db_game)
        BusyAsyncCall(self.get_service_installers, self.on_service_installers_loaded, db_game, update)

    def install_by_id(self, appid):
        """Installs a game given the appid for the game on this service."""
        db_game = ServiceGameCollection.get_game(self.id, appid)
        if not db_game:
            logger.error("No game %s found for %s", appid, self.id)
            return None
        return self.install(db_game)

    def on_service_installers_loaded(self, result, error):
        if error:
            raise error  # bounce this error off the backstop for default handling

        service_installers, db_game, existing_game = result

        # If an existing game was found, it may have been updated,
        # and it's not safe to fire this until we get here.
        if existing_game:
            GAME_UPDATED.fire(existing_game)

        if service_installers and db_game:
            application = Gio.Application.get_default()
            application.show_installer_window(service_installers, service=self, appid=db_game["appid"])

    def simple_install(self, db_game):
        """A simplified version of the install method, used when a game doesn't need any setup"""
        installer = self.generate_installer(db_game)
        configpath = write_game_config(db_game["slug"], installer["script"])
        game_id = add_game(
            name=installer["name"],
            runner=installer["runner"],
            slug=installer["game_slug"],
            directory=self.get_game_directory(installer),
            installed=1,
            installer_slug=installer["slug"],
            configpath=configpath,
            service=self.id,
            service_id=db_game["appid"],
        )
        return game_id

    def add_installed_games(self):
        """Services can implement this method to scan for locally
        installed games and add them to lutris.

        This runs on a worker thread, and must no trigger UI actions -
        so no emitting signals here.
        """

    def get_game_directory(self, _installer):
        """Specific services should implement this"""
        return ""

    def get_game_platforms(self, db_game: dict) -> List[str]:
        """Interprets the database record for this game from this service
        to extract its platform, or returns an empty list if this is not available."""
        return []

    def resolve_game_id(self, appid):
        db_game = get_game_for_service(self.id, appid)

        if db_game and db_game.get("id"):
            return str(db_game.get("id"))

        return None

    def get_service_db_game(self, game: Game):
        """Returns the row dictionary for the service-game corresponding to the game given, if any, or None."""
        if game.service == self.id and game.appid:
            return ServiceGameCollection.get_game(self.id, game.appid)
        return None

    def get_game_release_date(self, db_game: dict):
        """Services can implement this method so games that are not in the
        database can be sorted by release date (Year) based on service data."""

    def get_game_release_year(self, db_game: dict):
        """Returns game release year
        Does not have to be rewritten if get_game_release_date returns
        rfc/iso compatible date (YYYY-MM-DD)"""
        date = self.get_game_release_date(db_game)
        return date[:4] if date else ""


class OnlineService(BaseService):
    """Base class for online gaming services"""

    online = True
    cookies_path = NotImplemented
    cache_path = NotImplemented
    requires_login_page = False

    login_url = NotImplemented
    login_window_width = 390
    login_window_height = 500
    login_user_agent = settings.DEFAULT_USER_AGENT
    redirect_uris = NotImplemented

    @property
    def credential_files(self):
        """Return a list of all files used for authentication"""
        return [self.cookies_path]

    def login(self, parent=None):
        if self.client_installer and not self.is_launcher_installed():
            NoticeDialog(
                _(
                    "This service requires a game launcher. The following steps will install it.\n"
                    "Once the client is installed, you can login to %s."
                )
                % self.name
            )
            application = Gio.Application.get_default()
            application.show_lutris_installer_window(game_slug=self.client_installer)
            return
        logger.debug("Connecting to %s", self.name)
        dialog = WebConnectDialog(self, parent)
        dialog.run()

    @property
    def is_login_in_progress(self) -> bool:
        """Set to true if the login process is underway; the credential files may be created at this
        time, but that does not count as 'authenticated' until the login process is over. This is used
        by WebConnectDialog since it creates its cookies before the login is actually complete.

        This is recorded with a file in ~/.cache/lutris so it will persist across Lutris
        restarts, just as the credentials themselves do. For this reason, we need to allow
        the user to login again even when a login is in progress."""
        return self._get_login_in_progress_path().exists()

    @is_login_in_progress.setter
    def is_login_in_progress(self, in_progress: bool) -> None:
        path = self._get_login_in_progress_path()
        if in_progress:
            path.touch()
        else:
            path.unlink(missing_ok=True)

    def _get_login_in_progress_path(self) -> Path:
        return Path(os.path.join(settings.CACHE_DIR, f"{self.name}-login-in-progress"))

    def is_authenticated(self):
        """Return whether the service is authenticated"""
        if self.is_login_in_progress:
            logger.warning("Tried to get auth status while login in progress")
            return False
        return all(system.path_exists(path) for path in self.credential_files)

    def wipe_game_cache(self):
        """Wipe the game cache, allowing it to be reloaded"""
        if self.cache_path:
            logger.debug("Deleting %s cache %s", self.id, self.cache_path)
            if os.path.isdir(self.cache_path):
                shutil.rmtree(self.cache_path)
            elif system.path_exists(self.cache_path):
                os.remove(self.cache_path)
        super().wipe_game_cache()

    def logout(self):
        """Disconnect from the service by removing all credentials"""
        self.wipe_game_cache()
        for auth_file in self.credential_files:
            try:
                os.remove(auth_file)
            except OSError:
                logger.warning("Unable to remove %s", auth_file)
        logger.debug("logged out from %s", self.id)
        SERVICE_LOGOUT.fire(self)

    def load_cookies(self):
        """Load cookies from disk"""
        if not system.path_exists(self.cookies_path):
            logger.warning("No cookies found in %s, please authenticate first", self.cookies_path)
            return
        cookiejar = WebkitCookieJar(self.cookies_path)
        cookiejar.load()
        return cookiejar



"""
Discord Rich Presence Loader

This will enable DiscordRichPresenceClient if pypresence is installed.
Otherwise, it will provide a dummy client that does nothing


THIS MODULE IS UNMAINTAINTED AND WILL BE MARKED FOR DEPRECATION UNLESS SOMEONE VOLUNTEERS TO PROPERLY MAINTAIN IT.
"""

from lutris.util.discord.base import DiscordRPCNull

try:
    from lutris.util.discord.client import DiscordRichPresenceClient
except ImportError:
    # If PyPresence is not present, and ImportError will raise, so we provide dummy client
    client = DiscordRPCNull()
else:
    # PyPresence is present, so we provide the client
    client = DiscordRichPresenceClient()



# pylint: disable=no-member
import os
from gettext import gettext as _
from gettext import ngettext
from typing import Callable, Iterable, List

from gi.repository import GObject, Gtk

from lutris import settings
from lutris.database.games import get_game_by_field, get_games
from lutris.game import Game
from lutris.gui.dialogs import QuestionDialog
from lutris.gui.widgets.gi_composites import GtkTemplate
from lutris.gui.widgets.utils import get_required_main_window, get_widget_children
from lutris.util import datapath
from lutris.util.jobs import AsyncCall
from lutris.util.library_sync import LibrarySyncer
from lutris.util.log import logger
from lutris.util.path_cache import remove_from_path_cache
from lutris.util.strings import get_natural_sort_key, gtk_safe, human_size
from lutris.util.system import get_disk_size, is_removeable


@GtkTemplate(ui=os.path.join(datapath.get(), "ui", "uninstall-dialog.ui"))
class UninstallDialog(Gtk.Dialog):
    """A dialog to uninstall and remove games. It lists the games and offers checkboxes to delete
    the game files, and to remove from the library."""

    __gtype_name__ = "UninstallDialog"

    header_bar: Gtk.HeaderBar = GtkTemplate.Child()
    message_label: Gtk.Label = GtkTemplate.Child()
    uninstall_game_list: Gtk.ListBox = GtkTemplate.Child()
    cancel_button: Gtk.Button = GtkTemplate.Child()
    uninstall_button: Gtk.Button = GtkTemplate.Child()
    delete_all_files_checkbox: Gtk.CheckButton = GtkTemplate.Child()
    remove_all_games_checkbox: Gtk.CheckButton = GtkTemplate.Child()

    def __init__(self, parent: Gtk.Window, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.parent = parent
        self._setting_all_checkboxes = False
        self.games: List[Game] = []
        self.any_shared = False
        self.any_protected = False
        self.init_template()
        self.show_all()

    def get_game_removal_rows(self) -> List["GameRemovalRow"]:
        return get_widget_children(self.uninstall_game_list, GameRemovalRow)

    def add_games(self, game_ids: Iterable[str]) -> None:
        new_game_ids = set(game_ids) - set(g.id for g in self.games)
        new_games = [Game(game_id) for game_id in new_game_ids]
        new_games.sort(key=lambda g: get_natural_sort_key(g.name))
        self.games += new_games

        new_rows = []
        for game in new_games:
            row = GameRemovalRow(game)
            self.uninstall_game_list.add(row)
            new_rows.append(row)

        self.update_deletability()
        self.update_folder_sizes(new_games)
        self.update_subtitle()
        self.update_message()
        self.update_all_checkboxes()
        self.update_uninstall_button()
        self.uninstall_game_list.show_all()

        # Defer the connection until all checkboxes are updated
        for row in new_rows:
            row.connect("row-updated", self.on_row_updated)

    def update_deletability(self) -> None:
        """Updates the can_delete_files property on each row; adding new rows can set this on existing rows
        (they might no longer violate the 'can't delete shared directory' rule). This also sets flags that
        are used by later update methods, so this must be called first."""
        self.any_shared = False
        self.any_protected = False

        def is_shared(directory: str) -> bool:
            dir_users = set(str(g["id"]) for g in get_games(filters={"directory": directory, "installed": 1}))
            for g in self.games:
                dir_users.discard(g.id)
            return bool(dir_users)

        for row in self.get_game_removal_rows():
            game = row.game
            if game.is_installed and game.directory:
                if game.config and is_removeable(game.directory, game.config.system_config):
                    shared_dir = is_shared(game.directory)
                    self.any_shared = self.any_shared or shared_dir
                    row.can_delete_files = not shared_dir
                else:
                    row.can_delete_files = False
                    self.any_protected = True
            else:
                row.can_delete_files = False

    def update_folder_sizes(self, new_games: List[Game]) -> None:
        """Starts fetching folder sizes for new games added to the dialog; we only
        do this for the games given in 'new_games', however."""
        folders_to_size = []
        folders_seen = set()

        for row in self.get_game_removal_rows():
            game = row.game
            if game in new_games and game.is_installed and game.directory:
                if game.directory not in folders_seen:
                    folders_seen.add(game.directory)
                    folders_to_size.append(game.directory)
                row.show_folder_size_spinner()

        if folders_to_size:
            AsyncCall(
                self._get_next_folder_size,
                self._get_next_folder_size_cb,
                folders_to_size,
            )

    def update_subtitle(self) -> None:
        subtitle = self.build_subtitle()

        self.header_bar.set_subtitle(subtitle)

    def build_subtitle(self) -> str:
        """Updates the dialog subtitle according to what games are being removed."""
        to_uninstall = [g for g in self.games if g.is_installed]
        to_remove = [g for g in self.games if not g.is_installed]

        count_uninstall = len(to_uninstall)
        count_remove = len(to_remove)

        if not to_remove:
            if count_uninstall == 1:
                return _("Uninstall %s") % gtk_safe(to_uninstall[0].name)

            return ngettext("Uninstall %d game", "Uninstall %d games", count_uninstall) % count_uninstall

        if not to_uninstall:
            if count_remove == 1:
                return _("Remove %s") % gtk_safe(to_remove[0].name)

            return ngettext("Remove %d game", "Remove %d games", count_remove) % count_remove

        uninstall_part = ngettext("Uninstall %d game", "Uninstall %d games", count_uninstall) % count_uninstall
        remove_part = ngettext("remove %d game", "remove %d games", count_remove) % count_remove

        return _("%s and %s") % (uninstall_part, remove_part)

    def update_message(self) -> None:
        """Updates the message label at the top of the dialog."""
        to_uninstall = [g for g in self.games if g.is_installed]
        messages = []

        if to_uninstall:
            messages.append(_("After you uninstall these games, you won't be able play them in Lutris."))
            messages.append(
                _(
                    "Uninstalled games that you remove from the library will no longer appear in the "
                    "'Games' view, but those that remain will retain their playtime data."
                )
            )
        else:
            messages.append(_("After you remove these games, they will no longer appear in the 'Games' view."))

        if self.any_shared:
            messages.append(
                _(
                    "Some of the game directories cannot be removed because they are shared "
                    "with other games that you are not removing."
                )
            )

        if self.any_protected:
            messages.append(_("Some of the game directories cannot be removed because they are protected."))

        if messages:
            self.message_label.set_markup("\n\n".join(messages))
            self.message_label.show()
        else:
            self.message_label.hide()

    def on_row_updated(self, row) -> None:
        directory = row.game.directory
        if directory and row.can_delete_files:
            for r in self.get_game_removal_rows():
                if row != r and r.game.directory == directory and r.can_delete_files:
                    r.delete_files = row.delete_files

        self.update_all_checkboxes()

    def update_all_checkboxes(self) -> None:
        """Sets the state of the checkboxes at the button that are used to control all
        settings together. While we are actually updating these checkboxes en-mass,
        this method does nothing at all."""

        def update(checkbox, is_candidate, is_set):
            set_count = 0
            unset_count = 0
            for row in self.get_game_removal_rows():
                if is_candidate(row):
                    if is_set(row):
                        set_count += 1
                    else:
                        unset_count += 1

            checkbox.set_active(set_count > 0)
            checkbox.set_inconsistent(set_count > 0 and unset_count > 0)
            checkbox.set_visible((set_count + unset_count) > 1 and (set_count > 0 or unset_count > 0))

        if not self._setting_all_checkboxes:
            self._setting_all_checkboxes = True
            try:
                update(
                    self.delete_all_files_checkbox,
                    lambda row: row.can_delete_files,
                    lambda row: row.delete_files,
                )

                update(
                    self.remove_all_games_checkbox,
                    lambda row: True,
                    lambda row: row.remove_from_library,
                )
            finally:
                self._setting_all_checkboxes = False

    def update_uninstall_button(self) -> None:
        if any(g for g in self.games if g.is_installed):
            self.uninstall_button.set_label(_("Uninstall"))

    @GtkTemplate.Callback
    def on_delete_all_files_checkbox_toggled(self, _widget):
        def update_row(row, active):
            if row.can_delete_files:
                row.delete_files = active

        self._apply_all_checkbox(self.delete_all_files_checkbox, update_row)

    @GtkTemplate.Callback
    def on_remove_all_games_checkbox_toggled(self, _widget):
        def update_row(row, active):
            row.remove_from_library = active

        self._apply_all_checkbox(self.remove_all_games_checkbox, update_row)

    def _apply_all_checkbox(self, checkbox, row_updater: Callable[["GameRemovalRow", bool], None]):
        """Sets the state of the checkboxes on all rows to agree with 'checkbox';
        the actual change is performed by row_updater, so this can be used for
        either checkbox."""
        if not self._setting_all_checkboxes and checkbox.get_visible():
            active = checkbox.get_active()
            self._setting_all_checkboxes = True

            for row in self.get_game_removal_rows():
                row_updater(row, active)

            self._setting_all_checkboxes = False
            self.update_all_checkboxes()

    @GtkTemplate.Callback
    def on_cancel_button_clicked(self, _widget) -> None:
        self.destroy()

    @GtkTemplate.Callback
    def on_remove_button_clicked(self, _widget) -> None:
        rows = list(self.get_game_removal_rows())
        dirs_to_delete = list(set(row.game.directory for row in rows if row.delete_files))

        if dirs_to_delete:
            if len(dirs_to_delete) == 1:
                question = _("Please confirm.\nEverything under <b>%s</b>\nwill be moved to the trash.") % gtk_safe(
                    dirs_to_delete[0]
                )
            else:
                question = _("Please confirm.\nAll the files for %d games will be moved to the trash.") % len(
                    dirs_to_delete
                )

            dlg = QuestionDialog(
                {
                    "parent": self,
                    "question": question,
                    "title": _("Permanently delete files?"),
                }
            )

            if dlg.result != Gtk.ResponseType.YES:
                return

        library_sync_enabled = settings.read_bool_setting("library_sync_enabled", True)
        games_removed_from_library = []
        library_syncer = LibrarySyncer() if library_sync_enabled else None

        for row in rows:
            if library_syncer and row.remove_from_library:
                games_removed_from_library.append(get_game_by_field(row.game._id, "id"))
            row.perform_removal()

        if library_syncer and games_removed_from_library:

            def sync_local_library():
                library_syncer.sync_local_library()
                library_syncer.delete_from_remote_library(games_removed_from_library)

            AsyncCall(sync_local_library, None)

        get_required_main_window().on_game_removed()
        self.destroy()

    def on_response(self, _dialog, response: Gtk.ResponseType) -> None:
        if response in (
            Gtk.ResponseType.DELETE_EVENT,
            Gtk.ResponseType.CANCEL,
            Gtk.ResponseType.OK,
        ):
            self.destroy()

    @staticmethod
    def _get_next_folder_size(directories):
        """This runs on a thread and computes the size of the first directory in
        directories; the _get_next_folder_size_cb will run this again if required,
        until all directories have been sized."""
        directory = directories.pop(0)
        size = get_disk_size(directory)
        return directory, size, directories

    def _get_next_folder_size_cb(self, result, error):
        if error:
            logger.error(error)
            return

        directory, size, remaining_directories = result

        if remaining_directories:
            AsyncCall(
                self._get_next_folder_size,
                self._get_next_folder_size_cb,
                remaining_directories,
            )

        for row in self.get_game_removal_rows():
            if directory == row.game.directory:
                row.show_folder_size(size)


class GameRemovalRow(Gtk.ListBoxRow):
    __gsignals__ = {
        "row-updated": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, game: Game):
        super().__init__(activatable=False)
        self.game = game
        self._can_delete_files = False
        self.delete_files_checkbox: Gtk.CheckButton = None
        self.folder_size_spinner: Gtk.Spinner = None
        self.directory_label: Gtk.Label = None

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(hbox, False, False, 0)

        label = Gtk.Label(label=game.name, selectable=True)
        hbox.pack_start(label, False, False, 0)

        self.remove_from_library_checkbox = Gtk.CheckButton(label=_("Remove from Library"), halign=Gtk.Align.START)
        self.remove_from_library_checkbox.set_active(False)
        self.remove_from_library_checkbox.connect("toggled", self.on_checkbox_toggled)
        hbox.pack_end(self.remove_from_library_checkbox, False, False, 0)

        if game.is_installed and self.game.directory:
            self.delete_files_checkbox = Gtk.CheckButton(label=_("Delete Files"))
            self.delete_files_checkbox.set_sensitive(False)
            self.delete_files_checkbox.set_active(False)
            self.delete_files_checkbox.set_tooltip_text(self.game.directory)
            self.delete_files_checkbox.connect("toggled", self.on_checkbox_toggled)

            hbox.pack_end(self.delete_files_checkbox, False, False, 0)

            dir_box = Gtk.Box(
                orientation=Gtk.Orientation.HORIZONTAL,
                spacing=6,
                margin_left=6,
                margin_right=6,
                height_request=16,
            )
            self.directory_label = Gtk.Label(halign=Gtk.Align.START, selectable=True, valign=Gtk.Align.START)
            self.directory_label.set_markup(self._get_directory_markup())
            dir_box.pack_start(self.directory_label, False, False, 0)

            self.folder_size_spinner = Gtk.Spinner(visible=False, no_show_all=True)
            dir_box.pack_start(self.folder_size_spinner, False, False, 0)

            vbox.pack_start(dir_box, False, False, 0)
        self.add(vbox)

    def _get_directory_markup(self, folder_size: int = None):
        if not self.game.directory or not self.game.is_installed:
            return ""

        markup = gtk_safe(self.game.directory)
        if folder_size is not None:
            markup += f" <i>({human_size(folder_size)})</i>"
        return "<span font_desc='8'>%s</span>" % markup

    def on_checkbox_toggled(self, _widget):
        self.emit("row-updated")

    def show_folder_size_spinner(self):
        if self.folder_size_spinner:
            self.folder_size_spinner.start()
            self.folder_size_spinner.show()

    def show_folder_size(self, folder_size: int) -> None:
        """Called to stop the spinner and show the size of the game folder."""
        if self.directory_label:
            self.directory_label.set_markup(self._get_directory_markup(folder_size))

        if self.folder_size_spinner:
            self.folder_size_spinner.stop()
            self.folder_size_spinner.hide()

    @property
    def delete_files(self) -> bool:
        """True if the game files should be deleted."""
        return bool(
            self.game.is_installed
            and self.game.directory
            and self.delete_files_checkbox
            and self.delete_files_checkbox.get_active()
        )

    @delete_files.setter
    def delete_files(self, active: bool) -> None:
        self.delete_files_checkbox.set_active(active)

    @property
    def can_delete_files(self):
        return self._can_delete_files

    @can_delete_files.setter
    def can_delete_files(self, can_delete):
        if self._can_delete_files != can_delete and self.delete_files_checkbox:
            self._can_delete_files = can_delete
            self.delete_files_checkbox.set_sensitive(can_delete)
            self.delete_files_checkbox.set_active(can_delete)

    @property
    def remove_from_library(self) -> bool:
        """True if the game should be removed from the database."""
        return bool(self.remove_from_library_checkbox.get_active())

    @remove_from_library.setter
    def remove_from_library(self, active: bool) -> None:
        self.remove_from_library_checkbox.set_active(active)

    def perform_removal(self) -> None:
        """Performs the actions this row describes, uninstalling or deleting a game."""
        # We uninstall installed games, and delete games where self.remove_from_library is true
        if self.game.is_installed:
            remove_from_path_cache(self.game)
            if self.remove_from_library:
                self.game.uninstall(delete_files=self.delete_files)
                self.game.delete()
            else:
                self.game.uninstall(delete_files=self.delete_files)
        elif self.remove_from_library:
            self.game.delete()



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Iterable, Optional, cast

from gi.repository import GLib, GObject, Gtk

from lutris.gui.dialogs import display_error
from lutris.gui.widgets.utils import get_required_main_window
from lutris.util.log import logger

if TYPE_CHECKING:
    from lutris.game import Game

AnyCallable = Callable[..., Any]


def watch_game_errors(game_stop_result: Any, game: "Game" = None) -> Callable[[AnyCallable], AnyCallable]:
    """Decorator used to catch exceptions and send events instead of propagating them normally.
    If 'game_stop_result' is not None, and the decorated function returns that, this will
    send game-stop and make the game stopped as well. This simplifies handling cancellation.
    Also, if an error occurs and is emitted, the function returns this value, so callers
    can tell that the function failed.

    If you do not provide a game object directly, it is assumed to be in the first argument to
    the decorated method (which is 'self', typically).
    """
    captured_game = game

    def inner_decorator(function: AnyCallable) -> AnyCallable:
        @wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """Catch all exceptions and emit an event."""
            game = captured_game if captured_game else args[0]

            try:
                result = function(*args, **kwargs)
                if game_stop_result is not None and result == game_stop_result and game.state != game.STATE_STOPPED:
                    game.stop_game()
                return result
            except Exception as ex:
                logger.exception("%s has encountered an error: %s", game, ex, exc_info=ex)
                if game.state != game.STATE_STOPPED:
                    game.stop_game()
                game.signal_error(ex)
                return game_stop_result

        return cast(AnyCallable, wrapper)

    return inner_decorator


def _get_error_parent(error_objects: Iterable[Optional[GObject.Object]]) -> Gtk.Window:
    """Obtains a top-level window to use as the parent of an
    error, by examining s list of objects. Any that are None
    are skipped; we call get_toplevel() on each object that has
    this method, and return the first non-None result.

    If this fails, we turn to the application's main window instead."""

    for error_object in error_objects:
        if not error_object:
            continue

        try:
            if error_object and hasattr(error_object, "get_toplevel"):
                toplevel: Gtk.Window = error_object.get_toplevel()
                if toplevel:
                    return toplevel
        except GLib.GError:  # type:ignore
            pass  # hasattr() is always true for (some) GObjects, but the method fails when used

    return get_required_main_window()


def _create_error_wrapper(
    handler: AnyCallable, handler_name: str, error_result: Any, error_method_name: str, connected_object: Any = None
) -> AnyCallable:
    """Wraps a handler function in an error handler that will log and then report
    any exceptions, then return the 'error_result'."""

    handler_object = handler.__self__ if hasattr(handler, "__self__") else None

    def error_wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return handler(*args, **kwargs)
        except Exception as ex:
            logger.exception("Error handling %s: %s", handler_name, ex)

            if handler_object and hasattr(handler_object, error_method_name):
                error_method = getattr(handler_object, error_method_name)
                error_method(ex)
            else:
                display_error(ex, _get_error_parent([handler_object, connected_object]))
            return error_result

    return cast(AnyCallable, error_wrapper)


def init_exception_backstops() -> None:
    """This function is called once only, during startup, and replaces ("swizzles") a bunch of
    callback setup functions in GLib. The callbacks are all wrapped with error handlers that
    log the error and report it.

    This is important to do because PyGObject will straight up crash if an exception escapes
    these handlers; it's better to tell the user and try to survive.

    You can provide certain methods to provide error handling, but if you do not you get an
    ErrorDialog. Put these handling methods on the same object as the callback method itself.

    We take care of these methods:
        GObject.Object.connect (via on_signal_error(self, error))
        GObject.add_emission_hook (via on_emission_hook_error(self, error))
        GLib.idle_add (via on_idle_error(self, error))
        GLib.timeout_add (via on_timeout_error(self, error))

    Idle and timeout handlers will be disconnected if this happens to avoid repeated error reports,
    but signals and emission hooks will remain connected.
    """

    def _error_handling_connect(
        self: Gtk.Widget, signal_spec: str, handler: AnyCallable, *args: Any, **kwargs: Any
    ) -> Any:
        error_wrapper = _create_error_wrapper(
            handler,
            f"signal '{signal_spec}'",
            error_result=None,
            error_method_name="on_signal_error",
            connected_object=self,
        )
        return _original_connect(self, signal_spec, error_wrapper, *args, **kwargs)

    def _error_handling_add_emission_hook(
        emitting_type: GObject.Object, signal_spec: str, handler: AnyCallable, *args: Any, **kwargs: Any
    ) -> Any:
        error_wrapper = _create_error_wrapper(
            handler,
            f"emission hook '{emitting_type}.{signal_spec}'",
            error_result=True,  # stay attached
            error_method_name="on_emission_hook_error",
        )
        return _original_add_emission_hook(emitting_type, signal_spec, error_wrapper, *args, **kwargs)

    def _error_handling_idle_add(handler: AnyCallable, *args: Any, **kwargs: Any) -> Any:
        error_wrapper = _create_error_wrapper(
            handler,
            "idle function",
            error_result=False,  # stop calling idle func
            error_method_name="on_idle_error",
        )
        return _original_idle_add(error_wrapper, *args, **kwargs)

    def _error_handling_timeout_add(interval: int, handler: AnyCallable, *args: Any, **kwargs: Any) -> Any:
        error_wrapper = _create_error_wrapper(
            handler,
            "timeout function",
            error_result=False,  # stop calling timeout fund
            error_method_name="on_timeout_error",
        )
        return _original_timeout_add(interval, error_wrapper, *args, **kwargs)

    _original_connect = Gtk.Widget.connect
    GObject.Object.connect = _error_handling_connect

    _original_add_emission_hook: Callable[[GObject.Object, str, AnyCallable], Any] = GObject.add_emission_hook
    GObject.add_emission_hook = _error_handling_add_emission_hook

    _original_idle_add = GLib.idle_add
    GLib.idle_add = _error_handling_idle_add

    _original_timeout_add = GLib.timeout_add
    GLib.timeout_add = _error_handling_timeout_add



"""Class to manipulate a process"""

import os

from lutris.util.log import logger

IGNORED_PROCESSES = (
    "tracker-store",
    "tracker-extract",
    "kworker",
)


class Process:
    """Python abstraction a Linux process"""

    def __init__(self, pid):
        try:
            self.pid = int(pid)
        except ValueError as err:
            raise ValueError("'%s' is not a valid pid" % pid) from err

    def __repr__(self):
        return "Process {}".format(self.pid)

    def __str__(self):
        return "{} ({}:{})".format(self.name, self.pid, self.state)

    def _read_content(self, file_path):
        """Return the contents from a file in /proc"""
        try:
            with open(file_path, encoding="utf-8", errors="replace") as proc_file:
                content = proc_file.read()
        except PermissionError:
            return ""
        except (ProcessLookupError, FileNotFoundError) as ex:
            logger.debug(ex)
            return ""
        return content.strip("\x00")

    def get_stat(self, parsed=True):
        stat_filename = "/proc/{}/stat".format(self.pid)
        try:
            with open(stat_filename, encoding="utf-8", errors="replace") as stat_file:
                _stat = stat_file.readline()
        except (ProcessLookupError, FileNotFoundError):
            return None
        if parsed:
            return _stat[_stat.rfind(")") + 1 :].split()
        return _stat

    def get_thread_ids(self):
        """Return a list of thread ids opened by process."""
        basedir = "/proc/{}/task/".format(self.pid)
        if os.path.isdir(basedir):
            try:
                return os.listdir(basedir)
            except FileNotFoundError:
                return []
        else:
            return []

    def get_children_pids_of_thread(self, tid):
        """Return pids of child processes opened by thread `tid` of process."""
        children_path = "/proc/{}/task/{}/children".format(self.pid, tid)
        try:
            with open(children_path, encoding="utf-8", errors="replace") as children_file:
                children_content = children_file.read()
        except (FileNotFoundError, ProcessLookupError, PermissionError):
            children_content = ""
        return children_content.strip().split()

    @property
    def name(self):
        """Filename of the executable."""
        _stat = self.get_stat(parsed=False)
        if _stat:
            return _stat[_stat.find("(") + 1 : _stat.rfind(")")]
        return None

    @property
    def state(self):
        """One character from the string "RSDZTW" where R is running, S is
        sleeping in an interruptible wait, D is waiting in uninterruptible disk
        sleep, Z is zombie, T is traced or stopped (on a signal), and W is
        paging.
        """
        _stat = self.get_stat()
        if _stat:
            return _stat[0]
        return None

    @property
    def cmdline(self):
        """Return command line used to run the process `pid`."""
        cmdline_path = "/proc/{}/cmdline".format(self.pid)
        _cmdline_content = self._read_content(cmdline_path)
        if _cmdline_content:
            return _cmdline_content.replace("\x00", " ").replace("\\", "/")

    @property
    def cwd(self):
        """Return current working dir of process"""
        cwd_path = "/proc/%d/cwd" % int(self.pid)
        return os.readlink(cwd_path)

    @property
    def environ(self):
        """Return the process' environment variables"""
        environ_path = "/proc/{}/environ".format(self.pid)
        _environ_text = self._read_content(environ_path)
        if not _environ_text or "=" not in _environ_text:
            return {}
        env_vars = []
        for line in _environ_text.split("\x00"):
            if "=" not in line:
                continue
            env_vars.append(line.split("=", 1))
        return dict(env_vars)

    @property
    def children(self):
        """Return the child processes of this process"""
        _children = []
        for tid in self.get_thread_ids():
            for child_pid in self.get_children_pids_of_thread(tid):
                _children.append(Process(child_pid))
        return _children

    def iter_children(self):
        """Iterator that yields all the children of a process"""
        for child in self.children:
            yield child
            yield from child.iter_children()

    def wait_for_finish(self):
        """Waits until the process finishes
        This only works if self.pid is a child process of Lutris
        """
        try:
            pid, ret_status = os.waitpid(int(self.pid) * -1, 0)
        except OSError as ex:
            logger.error("Failed to get exit status for PID %s", self.pid)
            logger.error(ex)
            return -1
        logger.info("PID %s exited with code %s", pid, ret_status)
        return ret_status



import json
import os
import re
from configparser import ConfigParser
from gettext import gettext as _

from lutris import settings
from lutris.services.base import BaseService
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util import system
from lutris.util.strings import slugify

SCUMMVM_CONFIG_FILE = os.path.join(os.path.expanduser("~/.config/scummvm"), "scummvm.ini")


# Dummy banner. Maybe the download from lutris should be implemented at this place
class ScummvmBanner(ServiceMedia):
    service = "scummvm"
    source = "local"
    size = (96, 32)
    file_patterns = ["%s.png"]
    dest_path = settings.CACHE_DIR


class ScummvmService(BaseService):
    id = "scummvm"
    icon = "scummvm"
    name = _("ScummVM")
    local = True
    medias = {"icon": ScummvmBanner}

    def load(self):
        if not system.path_exists(SCUMMVM_CONFIG_FILE):
            return

        config = ConfigParser()
        config.read(SCUMMVM_CONFIG_FILE)
        config_sections = config.sections()

        for section in config_sections:
            if section == "scummvm":
                continue
            game = ScummvmGame()
            game.name = config[section]["description"]
            game.slug = slugify(re.split(r" \(.*\)$", game.name)[0])
            game.appid = section
            game.runner = "scummvm"
            game.lutris_slug = game.slug
            game.details = json.dumps({"path": config[section]["path"]})
            game.save()

    def generate_installer(self, db_game):
        details = json.loads(db_game["details"])
        return {
            "name": db_game["name"],
            "version": "ScummVM",
            "slug": db_game["slug"],
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "script": {
                "game": {
                    "game_id": db_game["appid"],
                    "path": details["path"],
                }
            },
        }

    def get_installed_slug(self, db_game):
        return db_game.get("lutris_slug") or slugify(db_game["name"])

    def get_installed_runner_name(self, db_game):
        return "scummvm"


class ScummvmGame(ServiceGame):
    service = "scummvm"
    runner = "scummvm"
